unit Donnees;

interface
uses
  Classes, SysUtils, FileUtil,GR32,GR32_Polygons,ComCtrls;
  type
    TTypeEchelle = (teLin,teLog);
    TBoolEtendu = (beNonDefini,beNonTeste,beTrue,beFalse); //beNonDefini si les paramètres ne coincident pas
    TTypeLien = (tlEt,TlOu,tlXor);
    TMaxiMini = Record
      Mini : real;
      Maxi : real;
      Couleur : TColor32;
    end;
    TParam = class(TObject)
      Nom : string;
      Alias : string;
      Numero : integer;
      TypeEchelle : TTypeEchelle;
      Puis10Log : integer;  //Nombre de décades
       // si Puis10Log > 0 , ValPourZero indique la valeur correspondant à zéro
      ValPourZero : real; // Si Puis10Log > 0 et  ValPourZero = 0 on prendra 1 pour la valeur 0
      Resolution : integer;
      Masque : LongWord;
      Nombrebits : integer;
      Sauvegarder : Boolean;
    end;
    TGraduations = class(TObject)
      PX : Array of integer;
      VX : Array of string;
      PY : Array of integer;
      VY : Array of string;
      YMax : integer; //Si histogramme
    end;
    TCategorie = class(TObject)
      Nom : string;
      AncienNom : string;
      FComposite : Boolean; // si composite constituée de sous-categories
                           // dans ce cas NbParams = 0
      CatAnnexe : Boolean; //categorie servant uniquement à définir une Composite ne pas colorer ni afficher en légende
      CatFille : TCategorie; //si CatAnnexe, la Categorie composée associée
      // pour les categories composites
      CatParents : Array of TCategorie;   // Parents directs
      CatAscendants : Array of TCategorie; // Tous les Ascendants
      Lien : TTypeLien;
      // pour les catégories simples
      NbParams : integer;
      Params : Array of TParam; //** voir le cas de plusieurs fichiers et des mêmes Catégories
      Resolution : integer;
      //si NbParams = 1;
      ValeurMin : real; //Valeurs
      ValeurMax : Real; //Valeurs
      // si NbParams = 2;
      PolygoneDef : TPolygon32;  //Valeurs entières correspondant aux valeurs réelles sans multiplier par 65535

      Inverser : Boolean; // True si Paramètres inversés dans PolygoneDef

      // pour toutes les catégories
      CatFilles : Array of TCategorie; //Descendants directs
      Active : Boolean;
      Couleur : TColor32;
      Resultat : TBoolEtendu;  // pour stoker le résultat d'un test (et le réutiliser pour les Catégories composées)
      Numero : integer;
      function TestePoint (LePoint: TFloatPoint;ResolutionD :integer):Boolean;
      function Inclut(LaDonnee : Array of single;LesParametres : Array of TParam;ResolutionD : integer) : TBoolEtendu;
      procedure SetNbParams (nombre : integer);
      procedure SetComposite (value : Boolean);
      function EstAscendantDe (Cat : TCategorie) : Boolean;
      Constructor Create;
      Destructor Destroy;Override;
    published
      property NombreParams : integer read NbParams write SetNbParams;
      property Composite : Boolean read FComposite write SetComposite;
    end;
    TGroupeCat = Array of TCategorie;
    TDonneesCyto = class;
    TEnsembleCategories = class(TObject)
      NbCategorieDeBase : integer;
      LesCategoriesDeBase : Array of TCategorie;
      LesCategoriesComp : Array of TCategorie;
      NbCategorieComp : integer;
      ParamsDeBase : Array of TParam;
      procedure RAZ;
      procedure ChercheDependantes(LaCategorieP : TCategorie;Var CatDep : TGroupeCat);
      procedure RAZResultats(Conserver : Boolean);//si Conserver conserve beNonDefini, sinon remet tout à beNonTeste
      procedure AjouteDeBase(LaCategorie : TCategorie);
      procedure AjouteComp(LaCategorie : TCategorie);
      function LaCategorie (LaCouleur : TColor32) : TCategorie;
      function LeNumero(UneCategorie : TCategorie) : integer;
      function ConserverCat(LaCategorie : TCategorie; Param1, Param2 : TParam;SerieDonnees:TDonneesCyto):Boolean;
      function RectangleLegende(LeBitmap : TBitmap32;Position : TFloatPoint; Param1, Param2 : TParam; SerieDonnees:TDonneesCyto;Var NbCol,NbLig : integer;Var lgnm,htnm :real) : TRect;
      procedure DessineLegende(LeBitmap : TBitmap32;Position : TFloatPoint; Param1, Param2 : TParam;SerieDonnees:TDonneesCyto; Var NbCol,NbLig : integer); // param <> nil pour les histogrammes
    end;
    TDonneesEvents = Array of Array of single;
    TDonneesCyto = class(TObject)
      Nom : string;
      Cytometre : string;
      NumSerie : string;
      NomFichier : String;
      NbEvenements : integer;
      FormatFichier : string;
      LaDate : string;
      Version : string;
      Projet : string;
      Cellules : string;
      ParamSerie : string; // paramètre qui différencie les différents jeux de données chargés (le plus souvent le temps)
      UniteParamSerie : string;
      ValeurParamSerie : integer;
      NbParam : integer; // nombre de paramètres
      OrdreBits : string;
      TypeDonnees : string[1]; //I pour entier, F pour réel single , D pour réel double, A pour AscII
      Parametres : Array of TParam;
      //** voir le cas du codage des données réelles (TFixed ?)
      //** voir le cas où la résolution est supérieure à 1024 en particulier pour les histogrammes et pour les graduations des axes
      Donnees : TDonneesEvents;//Array of Array of single; //Donnees[i,j] i : Evénement ; j : Paramètre
      Maxi : Array of real;
      Maxi2 : Array of Real;
      Couleurs : Array of TColor32; // une couleur par événement
      Resolution : integer;
      FacteurMult : real; // pour ajuster les valeurs, si la résolution a été modifiée NouvVal = ValOrig * FacteurMult
      MoysP : Array of real; // pour stocker les moyennes pour chaque paramètre
      EcTsP : Array of real; // pour stocker les ecart-types pour chaque paramètre
      MoysPCS : Array of Array of real; // pour stocker les moyennes des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      EcTsPCS : Array of Array of real; // pour stocker les ecart-types des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      NbEventCS : Array of integer;
      MoysPCC : Array of Array of real; // pour stocker les moyennes des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      EcTsPCC : Array of Array of real; // pour stocker les ecart-types des paramètres de chaque catégorie [i,j] i : categorie, j : paramètre
      NbEventCC : Array of integer;
      //NbParPixel : Array of Array of LongWord;
      NbParPixel : TBitmap32;
      DensiteMax : LongWord;

      procedure RazNbParPixel;
      procedure SetLengthNbParPixel(Nblg,Nbht : integer);
      function NumParam(LeParam : TParam) : integer;
      function NumParamTaille : integer;
      function NumParamComplexite : integer;

      procedure TraceHisto(LeBitmap : TBitmap32 ; LeParam : TParam;ValXEchLog, Corr : Boolean;CatRestr : TCategorie;Var MaxiX,MaxiY : real);
      // Param1 : Paramètre de l'abscisse, Param2 : celui de l'ordonnée. ValXEchLog,ValYEchLog : affichage log calculé en abscisse ou ordonnée
      //CatRestr : si <> nil limiter l'affichage aux points de la catégorie
      procedure TraceGraphe(LeBitmap : TBitmap32 ; Param1,Param2  : TParam ; Taillepoints : integer;ValXEchLog,ValYEchLog : Boolean;CatRestr : TCategorie;CoulDens : Boolean;Var MaxiX,MaxiY : real);
      procedure ChercheCouleurs(CatRestr : TCategorie;CoulDens : Boolean;Param1,Param2  : TParam);
      procedure CalculeStats;
      procedure CalculeDensite(largeur,hauteur : integer; NumParam1, NumParam2 : integer; MaXiY, dx, dy : real; Taillepoints : integer;ValXEchLog,ValYEchLog : Boolean);
      function EnregistreFCS(NouvNomFichier: string;CatRestr : TCategorie;ProgressBar:TProgressBar) : Boolean;
      function AjouteStreamFCS(CatRestr : TCategorie;ProgressBar:TProgressBar;Dernier : Boolean;Var MonFichier : TFileStream): Boolean;
      Constructor Cree(Nm:string;Nombre,Nbp : integer);
    end;
  Const
    qltt = 1;
    DP16 = 65536;
    clQuasi32 = 4294967294;
  Var
    LesDonneesCyto : Array of TDonneesCyto;
    LesParams : Array of TParam;
    LesCategories : TEnsembleCategories;
    MargeG,MargeD,MargeH,MargeB :integer;
    procedure ModifieSerie(SAM :TDonneesCyto;FacteurMult : real); //Les résolutions sont différentes, on ajuster les valeurs sur la plus grande
    procedure VerifieParametres;
    procedure RAZParametres;
    function LaCategorieHisto(Valeur : real;MonParametre : TParam;LaResolution : integer):TCategorie;
    function LaCategorieGraphe(LePoint : TPoint;Param1,Param2 : TParam;LaResolution : integer):TCategorie;
    function Compatible(LaCategorie : TCategorie ; SerieDonnees : TDonneesCyto) : Boolean;
    //function ChargeDuStream(LeStream: TStream; Nouvelle : Boolean) : Boolean;
    //function ChargeDuFichier(NomFichier: string; Nouvelle : Boolean) : Boolean;
    function ChargeDuFichierTexte(NomFichier: string; Nouvelle : Boolean) : Boolean;
    function PtInRect(Rectg : TRect;LePoint : TPoint) : Boolean;
    function TransposePolygone(LePolygone : TPolygon32;coefx,coefy: real; MaxiY : real;ValXEchLog,ValYEchLog : Boolean) : TPolygon32;
    function TransposePolygoneInv(LePolygone : TPolygon32;largeur,hauteur:integer;MaxiX,MaxiY : real;Inverser : Boolean;ValXEchLog,ValYEchLog : Boolean) : TPolygon32;
    function CalculeGraduations(LeBitmap : TBitmap32; maxiX,maxiY:real;param1,param2 : TParam;ValXEchLog,ValYEchLog : Boolean) : TGraduations;
    function ConversionLog(Valeur : real):real;
    function ConversionExp(Valeur : real) : real;
    function FracAleatoire(Valeur : Real) : real;
    function MessageOuiNon(Lignes : Array of string) : integer;
implementation
uses Math,GR32_Transforms,buttons,Forms,FormDialogOuinon,Dialogs;  //,SwapLibrary

function MessageOuiNon(Lignes : Array of string) : integer;
Var
  FDialogOuiNon : TFDialogOuiNon;
  i, nbl : integer;
  LargeurMaxi : integer;
begin
  Result := 0;
  FDialogOuiNon := TFDialogOuiNon.Create(nil);
  Try
    //FDialogOuiNon.Texte.Caption := TexteMessage;
    LargeurMaxi := 0;
    nbl := Length(Lignes);
    for i := 0 to nbl-1 do
      begin
        if LargeurMaxi < FDialogOuiNon.Texte.Canvas.TextWidth(Lignes[i]) then
          LargeurMaxi := FDialogOuiNon.Texte.Canvas.TextWidth(Lignes[i]);
        if i = (nbl -1) then
          FDialogOuiNon.Texte.Caption := FDialogOuiNon.Texte.Caption + Lignes[i]
        else
          FDialogOuiNon.Texte.Caption := FDialogOuiNon.Texte.Caption + Lignes[i]+chr(10)+chr(13);
      end;
    FDialogOuiNon.ClientHeight := 2*FDialogOuiNon.MargeV+nbl*FDialogOuiNon.HauteurLigne+FDialogOuiNon.Panel1.Height;
    FDialogOuiNon.ClientWidth := 2*FDialogOuiNon.MargeH+LargeurMaxi;
    Result := FDialogOuiNon.ShowModal;
  finally
    FDialogOuiNon.Free;
  end;
end;

function LaCategorieGraphe(LePoint : TPoint;Param1,Param2 : TParam;LaResolution : integer):TCategorie;
Var
  i : integer;
  MaCategorie : TCategorie;
  XCor,YCor,coeff  : real;
begin
  Result := nil;
  for i := LesCategories.NbCategorieDeBase-1 downto 0 do
    begin
      MaCategorie := LesCategories.LesCategoriesDeBase[i];
      if (MaCategorie.NbParams = 2) then   //MaCategorie.Active and
        begin
          if (MaCategorie.Params[0] = Param1) and (MaCategorie.Params[1] = Param2) then
            begin
              coeff := MaCategorie.Resolution/LaResolution;
              XCor := LePoint.x * coeff/DP16;
              YCor := LePoint.y * coeff/DP16;
              if PtInPolygon(Fixedpoint(XCor,YCor),MaCategorie.PolygoneDef.Points[0]) then
                begin
                  Result := MaCategorie;
                  Break;
                end;
            end
          else if (MaCategorie.Params[1] = Param1) and (MaCategorie.Params[0] = Param2) then
            begin
              coeff := MaCategorie.Resolution/LaResolution;
              XCor := LePoint.y * coeff/DP16;
              YCor := LePoint.x * coeff/DP16;
              if PtInPolygon(Fixedpoint(XCor,YCor),MaCategorie.PolygoneDef.Points[0]) then
                begin
                  Result := MaCategorie;
                  Break;
                end;
            end;
        end;
    end;
end;

function LaCategorieHisto(Valeur : real;MonParametre : TParam;LaResolution : integer):TCategorie;
Var
  i:integer;
  MaCategorie : TCategorie;
  ValeurCor : real;
begin
  Result := nil;
  for i := LesCategories.NbCategorieDeBase-1 downto 0 do
    begin
      MaCategorie := LesCategories.LesCategoriesDeBase[i];
      if (MaCategorie.NbParams = 1) and (MaCategorie.Params[0] = MonParametre) then   //MaCategorie.Active and
        begin
          ValeurCor := Valeur/LaResolution*MaCategorie.Resolution;
          if (ValeurCor >= MaCategorie.ValeurMin) and (ValeurCor <= MaCategorie.ValeurMax) then
            begin
              Result := MaCategorie;
              Break;
            end;
        end;
    end;
end;

function PtInRect(Rectg : TRect;LePoint : TPoint) : Boolean;
begin
  Result := (LePoint.x >= Rectg.Left) and (LePoint.x <= Rectg.Right)
    and (LePoint.y >= Rectg.Top) and (LePoint.Y <= Rectg.Bottom);
end;

procedure RAZParametres;
Var
  i : integer;
begin
  SetLength(LesParams,0);
end;

function CalculeGraduations(LeBitmap : TBitmap32;maxiX,maxiY:real;param1,param2: TParam;ValXEchLog,ValYEchLog : Boolean) : TGraduations;
Var
  i,n : integer;
  larg,haut : integer;
  coefX,coefY : real;
  GXM,GYM : integer;
  //ExPX,ExPY : integer;
  Mantissa: Double;
  Exponent: Integer;
  dx,dy : real;
  dxi,dyi:integer;
  nbGX,nbGY,ValI : integer;
  VMax : real;
  miniX,miniY : real;
begin
  larg := LeBitmap.Width-margeG-margeD;
  haut := LeBitmap.Height-margeH-margeB;
  Result := nil;
  if larg * haut >0 then
    begin
      coefX := larg/maxiX;  //-miniX
      coefY := haut/maxiY;  //-miniY
      if ValXEchLog then
        begin
          miniX :=200;
          CoefX :=larg/(maxiX); //-miniX
          VMax := PoWer(10,MaxiX/200);
          nbGX := Round(Log10(VMax))-1;
          Result := TGraduations.Create;
          GXM := 200*nbGX;
          dx := (GXM * coefX)/nbGX;
          SetLength(Result.PX,nbGX+1);
          SetLenGth(Result.VX,nbGX+1 );
          ValI := 10;
          for i := 0 to nbGX do
            begin
              if i = nbGX then
                Result.PX[i] :=margeG + round(i*dx)
              else
                Result.PX[i] :=margeG + round(i*dx);
              Result.VX[i] := IntToStr(ValI);
              ValI := 10 * ValI;
            end;
        end
      else if param1.TypeEchelle = TTypeEchelle.teLin then
        begin
          Exponent := Trunc(Log10(maxiX));
          Mantissa := PoWer(10,(Log10(maxiX)- Exponent));
          n := Trunc(Mantissa);
          GXM := Round(n*IntPower(10,Exponent));
          if n > 7 then
            nbGX := n div 2
          else if n > 3 then
            nbGX := n
          else if n> 1 then
            nbGX := 2*n
          else
            nbGX := 4*n;
          dxi := GXM div nbGX;
          dx := (GXM * coefX)/nbGX;
          Result := TGraduations.Create;
          SetLength(Result.PX,nbGX+1);
          SetLenGth(Result.VX,nbGX+1);
          for i := 0 to nbGX do
            begin
              Result.PX[i] :=margeG + Round(i*dx);
              Result.VX[i] := IntToStr(i*dxi);
            end;
        end
      else
        begin
          nbGX := param1.Puis10Log;
          dx := larg/nbGX;
          Result := TGraduations.Create;
          SetLength(Result.PX,nbGX+1);
          SetLenGth(Result.VX,nbGX+1);
          ValI := 1;
          for i := 0 to nbGX do
            begin
              Result.PX[i] :=margeG + round(i*dx);
              Result.VX[i] := IntToStr(ValI);
              ValI := 10 * ValI;
            end;
        end;
      if ValYEchLog then
        begin
          miniY :=200;
          CoefY :=haut/(maxiY);   //-miniY
          VMax := PoWer(10,MaxiY/200);
          nbGY := Round(Log10(VMax))-1;
          GYM := 200*nbGY;
          dy := (GYM * coefY)/nbGY;
          SetLength(Result.PY,nbGY+1);
          SetLenGth(Result.VY,nbGY+1 );
          ValI := 10;
          for i := 0 to nbGY do
            begin
              Result.PY[i] :=haut+margeH - Round(i*dy);
              Result.VY[i] := IntToStr(ValI);
              ValI := 10 * ValI;
            end;
        end
      else if (param2 = nil) or (param2.TypeEchelle = TTypeEchelle.teLin) then
        begin
          Exponent := Trunc(Log10(maxiY));
          Mantissa := PoWer(10,(Log10(maxiY)- Exponent));
          n := Trunc(Mantissa);
          GYM := Round(n*IntPower(10,Exponent));
          if n > 7 then
            nbGY := n div 2
          else if n > 3 then
            nbGY := n
          else if n>1 then
            nbGY := 2*n
          else
            nbGY := 4*n;
          dyi := GYM div nbGY;
          dy := (GYM * coefY)/nbGY;
          SetLength(Result.PY,nbGY+1);
          SetLenGth(Result.VY,nbGY+1);
          for i := 0 to nbGY do
            begin
              Result.PY[i] :=haut+margeH - Round(i*dy);
              Result.VY[i] := IntToStr(i*dyi);
            end;
        end
      else
        begin
          nbGY := Round(param2.Puis10Log);
          dY := haut/nbGY;
          SetLength(Result.PY,nbGY+1);
          SetLenGth(Result.VY,nbGY+1);
          ValI := 1;
          for i := 0 to nbGY do
            begin
              Result.PY[i] :=haut+margeH - Round(i*dy);
              Result.VY[i] := IntToStr(ValI);
              ValI := 10 * ValI;
            end;
        end;
    end;
end;
function TransposePolygone(LePolygone : TPolygon32;coefx,coefy : real; MaxiY : real;ValXEchLog,ValYEchLog : Boolean) : TPolygon32;
Var
  i : integer;
  X,Y : TFixed;
  Xp,Yp: TFixed;
  Xr,Yr : real;
 begin               //** vérifier ajustement si ValXEchLog ou/et ValYEchLog
  Result := TPolygon32.Create;
  Result.Assign(LePolygone);
  if (MaxiY > 0) then
    begin
      Result.Closed:= True;
      if Length(LePolygone.Points) > 0 then
        for i := 0 to Length(LePolygone.Points[0]) - 1 do
          begin
            Xp:=LePolygone.Points[0,i].X div DP16;
            Xr := (Xp-MargeG)*coefx;
            if ValXEchLog then
              X := Round(10*Power(10,Xr/200))
            else
              X := Round(Xr);
            Result.Points[0,i].X := X;
            Yp :=LePolygone.Points[0,i].Y div DP16;
            Yr := MaxiY-(Yp-MargeH)*coefy;
            if ValYEchLog then
              Y := Round(10*Power(10,Yr/200))
            else
              Y := Round(Yr);
            Result.Points[0,i].Y := Y
          end;
    end;
end;
function TransposePolygoneInv(LePolygone : TPolygon32;largeur,hauteur : integer;MaxiX,MaxiY : real;Inverser : Boolean;ValXEchLog,ValYEchLog : Boolean) : TPolygon32;
Var
  i : integer;
  coefX,coefY : real;    //** vérifier Echelle Log
  ValX, ValY : real;
begin
  Result := TPolygon32.Create;
  Result.Assign(LePolygone);
  if (MaxiX > 0) and (MaxiY > 0) then
    begin
      coefX := (largeur-margeG-margeD)/MaxiX;
      coefY := (hauteur-margeH-margeB)/MaxiY;
      if Length(LePolygone.Points) > 0 then
        begin
          for i := 0 to Length(LePolygone.Points[0]) - 1 do
            begin
              if Inverser then
                begin
                  ValX := LePolygone.Points[0,i].Y;
                  ValY := LePolygone.Points[0,i].X;
                end
              else
                begin
                  ValX := LePolygone.Points[0,i].X;
                  ValY := LePolygone.Points[0,i].Y;
                end;
              if ValXEchLog then
                begin
                  if ValX > 10 then
                    ValX := 200*(log10(ValX)-1)
                  else
                    ValX := 0;
                end;
              Result.Points[0,i].X := Round((MargeG + ValX * coefX)*DP16);
              if ValYEchLog then
                begin
                  if ValY > 10 then
                    ValY := 200*(log10(ValY)-1)
                  else
                    ValY := 0;
                end;
              Result.Points[0,i].Y := Round(((Hauteur-MargeB) - ValY * coefY)*DP16);
            end;
        end;
    end;
end;
procedure ModifieSerie(SAM :TDonneesCyto; FacteurMult : real); //Les résolutions sont différentes, on ajuster les valeurs sur la plus grande
Var
  i,j : integer;
begin
  SAM.FacteurMult := FacteurMult;
  for i := 0 to SAM.NbEvenements - 1 do
    for j := 0 to SAM.NbParam - 1 do
      begin
        SAM.Donnees[i,j] := Round(SAM.Donnees[i,j] * FacteurMult);
      end;
end;
procedure VerifieParametres; //Après ouverture d'un fichier
Var
  i,j,k : integer;
  NouvDonnees,SerieAModifier : TDonneesCyto;
  Differents : Boolean;
  P1,P2 : TParam;
  Nom1,Nom2,Alias1,Alias2 :string;
  FacteurMult : real;
begin
  if Length(LesDonneesCyto) > 1 then
    begin
      NouvDonnees := LesDonneesCyto[Length(LesDonneesCyto)-1];
      if NouvDonnees <> nil then
        begin
          for i := 0 to NouvDonnees.NbParam - 1 do
            begin
              for j :=0 to Length(LesParams) - 1 do
                begin
                  Differents := True;
                  P1 := NouvDonnees.Parametres[i];
                  P2 := LesParams[j];
                  Alias1 := UpperCase(P1.Alias);
                  Alias2 := UpperCase(P2.Alias);
                  if Alias1 = Alias2 then
                    begin
                      Nom1 := UpperCase(P1.Nom);
                      Nom2 := UpperCase(P2.Nom);
                      if (Nom1 = Nom2) and (P1.TypeEchelle = P2.TypeEchelle) and (P1.Puis10Log = P2.Puis10Log) then
                        begin //concordance
                          Differents := False;
                          if P1.Resolution <> P2.Resolution then  //Il faut ajuster
                            begin
                              if P2.Resolution > P1.Resolution then
                                begin
                                  NouvDonnees.Resolution := P2.Resolution;
                                  ModifieSerie(NouvDonnees,P2.Resolution/P1.Resolution);
                                end
                              else
                                begin
                                  FacteurMult := P1.Resolution/P2.Resolution;
                                  for k := 0 to Length(LesDonneesCyto) - 2 do
                                    begin
                                      SerieAModifier := LesDonneesCyto[k];
                                      if SerieAModifier.NumParam(P1) > -1 then  //si la série de Donnees contient le paramètre
                                        begin
                                          SerieAModifier.Resolution := P1.Resolution;
                                          ModifieSerie(SerieAModifier,FacteurMult);
                                        end;
                                    end;
                                  P2.Resolution := P1.Resolution;
                                end;
                            end;
                          NouvDonnees.Parametres[i] := P2; // on unifie les paramètres
                          //P1.Free; // on retirer le param en trop
                          Break;
                        end;
                    end;
                end;
              if Differents then  // On Ajoute à la liste
                begin
                  SetLength(LesParams,Length(LesParams)+1);
                  LesParams[Length(LesParams)-1] := NouvDonnees.Parametres[i];
                end;
            end;
        end;
    end
  else if Length(LesDonneesCyto) = 1 then   // début de constitution de la liste complète des paramètres
    begin
      SetLength(LesParams,LesDonneesCyto[0].NbParam);
      for i := 0 to LesDonneesCyto[0].NbParam - 1 do
        LesParams[i] := LesDonneesCyto[0].Parametres[i];
    end;
  // à la fin on réajuste les numéros des params
  for i := 0 to Length(LesParams) - 1 do
    LesParams[i].Numero := i;
end;
{function ChargeDuStream(LeStream: TStream; Nouvelle : Boolean) : Boolean;   //** n'est plus à jour en particulier lecture du param E $PnE
Var
  i,j,n,p : integer;
  LaDonneeCyto : TDonneesCyto;
  Tampon : AnsiChar;
  ValeurDW : LongWord;
  ValeurW : Word;
  ValeurB : Byte;
  ValeurF : single;
  taille : integer;
  Chaine, Chb : string;
  debutET, finET, debutD, finD,longET : integer;
  EnTete : TStringList;
  UnParam : TParam;
  Value4     :  LongWord;
  ShortFloat :  single  absolute Value4;  // Memory overlay trick

  procedure SauteLesBlancs;
  begin
    while (ord(Tampon)= $20) and (LeStream.Position < taille) do
      LeStream.Read(Tampon,1);
  end;
  function LitOctets(nbo:integer) : string;
  Var
    ii : integer;
  begin
    Result := '';
    ii:= 0;
    while (ii < nbo) and (LeStream.Position < taille) do
      begin
        LeStream.Read(Tampon,1);
        Result := Result + Tampon;
        inc(ii)
      end;
  end;
  function Litchaine : string;
  begin
    Result := '';
    while (ord(Tampon)<> $20) and (LeStream.Position < taille) do
      begin
        Result := Result + Tampon;
        LeStream.Read(Tampon,1);
      end;
  end;
  function LitValeurC(S : string) : integer;
  begin
      try
        Result := StrToInt(Trim(S));
      except
        Result :=0;
      end;
  end;
  function LitValeurF : integer;
  begin
      Chaine := LitChaine;
      result := LitValeurC(Chaine);
  end;
  function LitParamT(code :string): string;
  begin
    n := EnTete.IndexOf(code);
    if n > -1 then
      Result := EnTete[n+1]
    else
      Result := '';
  end;
  function LitParamN(code :string): integer;
  begin
    n := EnTete.IndexOf(code);
    if n > -1 then
      Result := LitValeurC(EnTete[n+1])
    else
      Result := 0;
  end;
  function LitParamR(code :string): real; //** voir les problèmes de représentation de la virgule
  begin
    n := EnTete.IndexOf(code);
    if n > -1 then
      begin
        try
          Result := StrToFloat(EnTete[n+1]);
        except
          Result := 0;
        end;
      end
    else
      Result := 0;
  end;
begin
  Result := False;
  if (LeStream <> Nil) and (Lestream.Size > 0) then
    begin
      taille := Lestream.Size;
      n := Length(LesDonneesCyto);
      if (n = 0) or Nouvelle then
        begin
          SetLength(LesDonneesCyto,n+1);
          LaDonneeCyto := TDonneesCyto.Create;
          LesDonneesCyto[n] := LaDonneeCyto;
        end
      else
        LaDonneeCyto := LesDonneesCyto[n-1];
      // Lecture de l'en-tête
      LaDonneeCyto.FormatFichier := '';
      LeStream.Seek(0,soBeginning);//se place au début du flux;
      LeStream.Read(Tampon,1); //lecture du Format de fichier
      LaDonneeCyto.FormatFichier := LitChaine;
      LeStream.Position :=10;
      Chb := Trim(LitOctets(8));
      try
        debutET := StrToInt(Chb);
      except
        debutET := 256;
      end;
      LeStream.Position :=18;
      Chb := Trim(LitOctets(8));
      try
        finET := StrToInt(Chb);
      except
        finET := 256;
      end;
      longET := finET-debutET+1;
      LeStream.Position :=26;   //début des données
      Chb := Trim(LitOctets(8));
      try
        debutD := StrToInt(Chb);
      except
        debutD := 0;
      end;
      LeStream.Position :=34;
      Chb := Trim(LitOctets(8));
      try
       finD := StrToInt(Chb);
      except
       finD := 0;
      end;

      LeStream.Position := DebutET;
      if (longET > 0) and (LeStream.Position < taille) then
        begin
          EnTete := TStringList.Create;
          try
            EnTete.StrictDelimiter :=True;
            Chb := LitOctets(longET);

            EnTete.Delimiter :=Chb[1];
            EnTete.DelimitedText := Chb;
             // nombre de paramètres
            p := LitParamN('$PAR');
            if p > 0 then
              begin
                LaDonneeCyto.NbParam := p;
                SetLength(LaDonneeCyto.Parametres,LaDonneeCyto.NbParam);
                // ** Lecture de $DATATYPE  (I pour Integer, F pour single precision IEEE floating point,
                // ** D pour double precision IEEE floating point, A pour ASCII
                chb := LitParamT('$DATATYPE');
                LaDonneeCyto.TypeDonnees := chb[1];
                chb := LitParamT('$DATATYPE');
                LaDonneeCyto.TypeDonnees := chb[1];
                chb := LitParamT('$MODE'); //L pour Linéaire

                if (chb[1] = 'L') and ((LaDonneeCyto.TypeDonnees = 'I') or (LaDonneeCyto.TypeDonnees = 'F')) then
                  begin
                    p := LitParamN('$TOT');
                    if p > 0 then
                      begin
                        LaDonneeCyto.NbEvenements := p;
                        LaDonneeCyto.LaDate := LitParamT('$DATE');
                        LaDonneeCyto.OrdreBits := LitParamT('$BYTEORD');
                        LADonneeCyto.Cytometre := LitParamT('$CYT');
                        SetLength(LaDonneeCyto.Maxi,LaDonneeCyto.NbParam);
                        SetLength(LaDonneeCyto.Maxi2,LaDonneeCyto.NbParam);
                        for i := 1 to LaDonneeCyto.NbParam do
                          begin
                            UnParam := TParam.Create;
                            UnParam.Numero := i-1;
                            UnParam.Nom :=LitParamT('$P'+IntToStr(i)+'N');
                            UnParam.Alias :=LitParamT('$P'+IntToStr(i)+'S');
                            UnParam.Nombrebits := LitParamN('$P'+IntToStr(i)+'B');
                            //UnParam.Puis10Log := LitParamR('$P'+IntToStr(i)+'E');
                            if UnParam.Puis10Log > 0 then
                              UnParam.TypeEchelle := teLog
                             else
                               UnParam.TypeEchelle := teLin;
                             UnParam.Resolution := LitParamN('$P'+IntToStr(i)+'R');
                             UnParam.Masque:=UnParam.Resolution - 1;
                             LaDonneeCyto.Parametres[i-1] := UnParam;
                             LaDonneeCyto.Resolution := UnParam.Resolution;
                             LaDonneeCyto.Maxi[i] := 0;
                             LaDonneeCyto.Maxi[i] := 0;
                          end;
                        LeStream.Position :=debutD;
                        SetLength(LaDonneeCyto.Donnees,LaDonneeCyto.NbEvenements);
                        SetLength(LaDonneeCyto.Couleurs,LaDonneeCyto.NbEvenements);
                        ValeurB :=0;
                        for i := 0 to LaDonneeCyto.NbEvenements-1 do
                          begin
                            SetLength(LaDonneeCyto.Donnees[i],LaDonneeCyto.NbParam);
                            LaDonneeCyto.Couleurs[i] := $FF000000;
                            for j := 0 to LaDonneeCyto.NbParam-1 do
                              if LeStream.Position < FinD then
                                begin
                                  if LaDonneeCyto.TypeDonnees = 'I' then  // valeurs entières
                                    begin
                                      if LaDonneeCyto.Parametres[j].Nombrebits = 8 then  // 8 bits
                                        begin
                                          LeStream.Read(ValeurB,1);
                                          LaDonneeCyto.Donnees[i,j] :=ValeurB;  // Donnees[i,j] i : Evénement ; j : Paramètre
                                        end
                                      else if LaDonneeCyto.Parametres[j].Nombrebits = 16 then  // 16 bits
                                        begin
                                          LeStream.Read(ValeurB,1);
                                          ValeurW :=ValeurB * 256;
                                          LeStream.Read(ValeurB,1);
                                          ValeurW :=ValeurW + ValeurB;
                                          ValeurW := ValeurW and LaDonneeCyto.Parametres[j].Masque;
                                          LaDonneeCyto.Donnees[i,j] :=ValeurW;  // Donnees[i,j] i : Evénement ; j : Paramètre
                                        end
                                      else if LaDonneeCyto.Parametres[j].Nombrebits = 32 then   // 32 bits
                                        begin
                                          LeStream.Read(ValeurB,1);
                                          ValeurW :=ValeurB * 256;
                                          LeStream.Read(ValeurB,1);
                                          ValeurW :=ValeurW + ValeurB;
                                          ValeurDW :=ValeurW * 65536;
                                          LeStream.Read(ValeurB,1);
                                          ValeurW :=ValeurB * 256;
                                          LeStream.Read(ValeurB,1);
                                          ValeurW :=ValeurW + ValeurB;
                                          ValeurDW := ValeurDW + ValeurW;
                                          ValeurDW := ValeurDW and LaDonneeCyto.Parametres[j].Masque;
                                          LaDonneeCyto.Donnees[i,j] :=ValeurDW;  // Donnees[i,j] i : Evénement ; j : Paramètre
                                        end;
                                      //** tenir compte des  cas 64 bits  ?
                                    end
                                  else if LaDonneeCyto.TypeDonnees = 'F' then   // valeurs réelles
                                    begin
                                      LeStream.Read(Value4,4);
                                      Value4 := Swap4(Value4);
                                      LaDonneeCyto.Donnees[i,j] :=ShortFloat;
                                    end;
                                  if LaDonneeCyto.Donnees[i,j] > LaDonneeCyto.Maxi[j] then
                                    LaDonneeCyto.Maxi[j] := LaDonneeCyto.Donnees[i,j]
                                  else if (LaDonneeCyto.Donnees[i,j] < LaDonneeCyto.Maxi[j]) and (LaDonneeCyto.Donnees[i,j] > LaDonneeCyto.Maxi2[j]) then   //LaDonneeCyto.Maxi[j]*0.5
                                    LaDonneeCyto.Maxi2[j] := LaDonneeCyto.Donnees[i,j];

                                end;
                          end;
                        VerifieParametres;
                        Result := True;
                      end;
                  end;
              end
            else
              MessageDlg('Le logiciel ne lit pas ce type de fichier',mtError,[mbOK],0);
          finally
            EnTete.Free;
          end;
        end;
    end;
end;
}
function ChargeDuFichierTexte(NomFichier: string; Nouvelle : Boolean) : Boolean;
Var
  i,j,n : integer;
  FicT,Ligne : TStringList;
  LaDonneeCyto : TDonneesCyto;
  NombreBits : integer;
  Resolution : integer;
  nombre : integer;
  UnParam : TParam;
  function ExtraitNombre(Symbole, Texte : string) : integer;
  Var
    p : integer;
  begin
    Result := - 1;
    p := pos(Symbole,Trim(Texte));
    try
      Result := StrToInt(copy(Trim(Texte),p+Length(Symbole),Length(Texte)));
    except

    end;
  end;
begin
  Result := False;
  FicT := TStringList.Create;
  Ligne := TStringList.Create;
  try
    FicT.LoadFromFile(NomFichier);
    if FicT.Count > 10 then
      begin
        n := Length(LesDonneesCyto);
        if (n = 0) or Nouvelle then
          begin
            SetLength(LesDonneesCyto,n+1);
            LaDonneeCyto := TDonneesCyto.Create;
            LesDonneesCyto[n] := LaDonneeCyto;
          end
        else
          LaDonneeCyto := LesDonneesCyto[n-1];

        LaDonneeCyto.NomFichier := NomFichier;
        Resolution := ExtraitNombre('Resolution:',FicT[0]);
        LaDonneeCyto.Resolution := Resolution;
        if Resolution <= 256 then
          NombreBits := 1
        else
          NombreBits := 2;
        Ligne.Delimiter := chr(9);
        Ligne.StrictDelimiter := True;
        Ligne.DelimitedText :=FicT[1];
        Nombre := ExtraitNombre('Events:',Ligne[0]);
        if Nombre > - 1 then
          LaDonneeCyto.NbEvenements := Nombre
        else
          begin
            exit;
          end;
        Nombre := ExtraitNombre('Parameters:',Ligne[1]);
        if Nombre > - 1 then
          LaDonneeCyto.NbParam := Nombre
        else
          begin
            exit;
          end;
        Ligne.DelimitedText :=FicT[3];
        if Ligne.Count = LaDonneeCyto.NbParam then
          begin
            SetLength(LaDonneeCyto.Parametres,LaDonneeCyto.NbParam);
            for i := 0 to Ligne.Count - 1 do
              begin
                UnParam := TParam.Create;
                LaDonneeCyto.Parametres[i] := UnParam;
                UnParam.Nom := Ligne[i];
                UnParam.Resolution := Resolution;
                UnParam.Nombrebits := NombreBits;
              end;
            Ligne.DelimitedText :=FicT[2];
            if Ligne.Count = LaDonneeCyto.NbParam then
              begin
                for i := 0 to Ligne.Count - 1 do
                  if Ligne[i] = 'Lin' then
                    LaDonneeCyto.Parametres[i].TypeEchelle := teLin
                  else if (pos('Log',Ligne[i]))> 0 then
                    begin
                      LaDonneeCyto.Parametres[i].TypeEchelle := teLog;
                      LaDonneeCyto.Parametres[i].Puis10Log := 4;
                    end;
              end;
            Ligne.DelimitedText :=FicT[4];
            if Ligne.Count = LaDonneeCyto.NbParam then
              begin
                for i := 0 to Ligne.Count - 1 do
                  LaDonneeCyto.Parametres[i].Alias := Ligne[i];
                // Saisie des données
                SetLength(LaDonneeCyto.Donnees,FicT.Count - 5);
                LaDonneeCyto.NbEvenements := FicT.Count - 5;
                SetLength(LaDonneeCyto.Couleurs,LaDonneeCyto.NbEvenements);

                for i := 5 to FicT.Count - 1 do
                  begin
                    Ligne.DelimitedText :=FicT[i];
                    SetLength(LaDonneeCyto.Donnees[i-5],LaDonneeCyto.NbParam);
                    LaDonneeCyto.Couleurs[i-5] := $FF000000;

                    if Ligne.Count >= LaDonneeCyto.NbParam then
                      begin
                        for j:= 0 to LaDonneeCyto.NbParam - 1 do
                          LaDonneeCyto.Donnees[i-5,j]:=StrToInt(Ligne[j]);
                      end;
                  end;
                VerifieParametres;
                Result := True;
              end;
          end;
      end;
  finally
    Fict.Free;
    Ligne.Free;
  end;
end;
{function ChargeDuFichier(NomFichier: string; Nouvelle : Boolean) : Boolean;
Var
  LeStream : TFileStream;
begin
  Result := False;
  if FileExists(NomFichier) then
    begin
      if UpperCase(ExtractFileExt(NomFichier)) = '.TXT' then
        begin
          Result := ChargeduFichierTexte(NomFichier,Nouvelle);
          if Result then
            LesDonneesCyto[Length(LesDonneesCyto)-1].NomFichier := Nomfichier;
        end
      else
        begin
          LeStream := TFileStream.Create(NomFichier,fmOpenRead or fmShareDenyNone);
          try
            Result := ChargeDuStream(LeStream, Nouvelle);
            if Result then
              LesDonneesCyto[Length(LesDonneesCyto)-1].NomFichier := Nomfichier;
          finally
            LeStream.Free;
          end;
        end;
    end;
end;
}
function Compatible(LaCategorie : TCategorie ; SerieDonnees : TDonneesCyto) : Boolean;
Var
  i : integer;
  CtPar1, CtPar2 : TParam;
begin
  if LaCategorie.Composite then
    begin
       // vérifier aussi si les params des ancêtres de la catégorie sont bien dans les données
      Result := True;
      for i := 0 to Length(LaCategorie.CatAscendants)-1 do
        begin
          CtPar1 := LaCategorie.CatAscendants[i].Params[0];
          CtPar2 := LaCategorie.CatAscendants[i].Params[1];
          if ((CtPar1 <> nil)and(SerieDonnees.NumParam(CtPar1) = -1)) or ((CtPar2 <> nil)and(SerieDonnees.NumParam(CtPar2) = -1)) then
            Result := False;
        end;
    end
  else
    begin
      Result := False;
      CtPar1 := LaCategorie.Params[0];
      CtPar2 := LaCategorie.Params[1];
      if (SerieDonnees.NumParam(CtPar1)> -1) and ((CtPar2 = nil) or (SerieDonnees.NumParam(CtPar2)> -1)) then
        Result :=True;
    end;
end;

// *** TEnsembleCategories ***
procedure TEnsembleCategories.RAZ;
Var
  i : integer;
  CtPar1, CtPar2 : TParam;
begin
  for i := NbCategorieDeBase - 1 downto 0 do
    LesCategoriesDeBase[i].Free;
  NbCategorieDeBase := 0;
  SetLength(LesCategoriesDeBase,NbCategorieDeBase);
  for i := NbCategorieComp - 1 downto 0 do
    LesCategoriesComp[i].Free;
  NbCategorieComp := 0;
  SetLength(LesCategoriesComp,NbCategorieComp);
  SetLength(ParamsDeBase,0);
end;
function TEnsembleCategories.ConserverCat(LaCategorie: TCategorie; Param1, Param2 : TParam;SerieDonnees:TDonneesCyto):Boolean;
Var
  i : integer;
  CtPar1, CtPar2 : TParam;
begin
  Result := False;
  if LaCategorie.Active and (Not LaCategorie.CatAnnexe) then
    begin
      if LaCategorie.Composite then
        begin
           // vérifier aussi si les params des ancêtres de la catégorie sont bien dans les données
          Result := True;
          for i := 0 to Length(LaCategorie.CatAscendants)-1 do
            begin
              CtPar1 := LaCategorie.CatAscendants[i].Params[0];
              CtPar2 := LaCategorie.CatAscendants[i].Params[1];
              if ((CtPar1 <> nil)and(SerieDonnees.NumParam(CtPar1) = -1)) or ((CtPar2 <> nil)and(SerieDonnees.NumParam(CtPar2) = -1)) then
                Result := False;
            end;
        end
      else
        begin
          CtPar1 := LaCategorie.Params[0];
          CtPar2 := LaCategorie.Params[1];
          if (SerieDonnees.NumParam(CtPar1)> -1) and ((CtPar2 = nil) or (SerieDonnees.NumParam(CtPar2)> -1)) then
            begin
              if ((param1 = nil) and (param2 = nil)) or ((param2 = nil) and (LaCategorie.NbParams = 1) and (CtPar1 = Param1)) then
                Result := True
              else if (param1 <> nil) and (param2 <> nil) and
                ((param1 = CtPar1) and (param2 = CtPar2)) or ((param2 = CtPar1) and (param1 = CtPar2)) then
                Result := True;
            end;
        end;
    end;
end;

function TEnsembleCategories.RectangleLegende(LeBitmap : TBitmap32;Position : TFloatPoint; Param1, Param2 : TParam;SerieDonnees:TDonneesCyto; Var NbCol,NbLig : integer;Var lgnm,htnm :real) : TRect;
Var
  i,n,j : integer;
  LeNom : string;
  largeur, hauteur : real;
  CtPar1, CtPar2 : TParam;
begin
  n := 0;
  lgnm := 0;
  htnm := 0;
  Result := Rect(0,0,0,0);
  for i := 0 to NbCategorieDeBase - 1 do
    begin
      if ConserverCat(LesCategoriesDeBase[i],Param1, Param2,SerieDonnees) then
{      if LesCategoriesDeBase[i].Active and (Not LesCategoriesDeBase[i].CatAnnexe) then
        begin
          CtPar1 := LesCategoriesDeBase[i].Params[0];
          CtPar2 := LesCategoriesDeBase[i].Params[1];
          if ((param1 = nil) and (param2 = nil)) or ((param2 = nil) and (LesCategoriesDeBase[i].NbParams = 1) and (CtPar1 = Param1)) then
}
            begin
              LeNom := LesCategoriesDeBase[i].Nom;
              inc(n);
              if lgnm < LeBitmap.TextWidth(LeNom) then
                lgnm := LeBitmap.TextWidth(LeNom);
              if htnm < LeBitmap.TextHeight(LeNom) then
                htnm := LeBitmap.TextHeight(LeNom);
            end;
{          else if (param1 <> nil) and (param2 <> nil) then
            begin
              if ((param1 = CtPar1) and (param2 = CtPar2)) or ((param2 = CtPar1) and (param1 = CtPar2)) then
                begin
                  LeNom := LesCategoriesDeBase[i].Nom;
                  inc(n);
                  if lgnm < LeBitmap.TextWidth(LeNom) then
                    lgnm := LeBitmap.TextWidth(LeNom);
                  if htnm < LeBitmap.TextHeight(LeNom) then
                    htnm := LeBitmap.TextHeight(LeNom);
                end;
            end;
        end;}
    end;
  //if (param1 = nil) and (param2 = nil) then
    for i := 0 to NbCategorieComp - 1 do
      begin
        if ConserverCat(LesCategoriesComp[i],Param1, Param2,SerieDonnees) then
          begin
            LeNom := LesCategoriesComp[i].Nom;
            inc(n);
            if lgnm < LeBitmap.TextWidth(LeNom) then
              lgnm := LeBitmap.TextWidth(LeNom);
            if htnm < LeBitmap.TextHeight(LeNom) then
              htnm := LeBitmap.TextHeight(LeNom);
          end;
      end;
  if (param1 <> nil) and (param2 <> nil) then  // couleurs par densité
    begin
      LeNom := 'Densité -  +';
      if lgnm < (LeBitmap.TextWidth(LeNom)+20-2*htnm+3) then
        lgnm := LeBitmap.TextWidth(LeNom)+20-2*htnm+3;
      if htnm < LeBitmap.TextHeight(LeNom) then
        htnm := LeBitmap.TextHeight(LeNom);
    end;
  if (n > 0) or ((param1 <> nil) and (param2 <> nil)) then
    begin
      if NbCol = 0 then
        NbCol := (n div 5)+1
      else if (n div NbCol) = 0 then
        NbCol := (n div 5)+1;
      NbLig := Ceil(n/NbCol);
      if (param1 <> nil) and (param2 <> nil) then
        inc(NbLig);
      Largeur := 2 +NbCol*(lgnm+2*htnm-3);
      hauteur := 4 + NbLig *(htnm+2);
      Result := Rect(Round(Position.X),Round(Position.Y),Round(Position.X+largeur),Round(Position.Y+hauteur));
    end;
end;
procedure TEnsembleCategories.DessineLegende(LeBitmap : TBitmap32;Position : TFloatPoint;Param1, Param2 : TParam;SerieDonnees:TDonneesCyto; Var NbCol,NbLig : integer); // param <> nil pour les histogrammes
 //param1 = nil pour les histogrammes seulement
Var
  i,n : integer;
  lgnm : real;
  htnm : real;
  LargeurCol : real;
  NumCol,NumLig : integer;
  RectLeg : TRect;
  CtPar1, CtPar2 : TParam;
  X, Y1, Y2 : integer;

  procedure EcritLeg(Texte : string;Couleur :TColor32;ModeTrait : Boolean);
  Var
    decCol,decLig : integer;
    RectTp : TRect;
  begin
    decCol := 4+Round(Position.X+NumCol*LargeurCol);
    decLig := 2+Round(Position.Y+NumLig*(htnm+2));
    LeBitmap.RenderText(decCol,decLig,Texte,qltt,clBlack32);
    RectTp.Left := Round(decCol + lgnm + htnm-8);
    if ModeTrait then
      RectTP.Top := Round(decLig+htnm-3)
    else
      RectTP.Top := Round(decLig+3);
    RectTP.Right := Round(decCol+lgnm+2*htnm-12);
    RectTP.Bottom := Round(decLig+htnm-1);
    LeBitmap.FillRectS(RectTP,Couleur);
    inc(n);
    inc(NumLig);
    if NumLig >= NbLig then
      begin
        NumLig := 0;
        inc(NumCol);
      end;
  end;
begin
  if (NbCategorieDeBase > 0) or ((param1 <> nil) and (param2 <> nil)) then
    begin
      // Comptage du nombre de lignes nécessaires  Mesure de la longueur maxi du texte  Délimitation du rectangle
      RectLeg := RectangleLegende(LeBitmap,Position,Param1,Param2,SerieDonnees,NbCol,NbLig,lgnm,htnm);
      LeBitmap.BeginUpdate;
      LeBitmap.FillRectS(RectLeg,clWhite32);
      LeBitmap.FrameRectS(RectLeg,clBlack32);
      n := 0;
      NumCol := 0;
      NumLig := 0;
      LargeurCol := lgnm + Round(2 * htnm) - 4;
      for i := 0 to NbCategorieDeBase - 1 do
        begin
          if ConserverCat(LesCategoriesDeBase[i],Param1, Param2,SerieDonnees) then
{          CtPar1 := LesCategoriesDeBase[i].Params[0];
          CtPar2 := LesCategoriesDeBase[i].Params[1];
          if LesCategoriesDeBase[i].Active and (Not LesCategoriesDeBase[i].CatAnnexe) then}
            if param2 = nil then
              EcritLeg(LesCategoriesDeBase[i].Nom,LesCategoriesDeBase[i].Couleur,False)
            else if ((param1 <> nil) and (param2 <> nil)) then  // cas d'un diagramme nuage de points avec couleurs liées à la densité
              EcritLeg(LesCategoriesDeBase[i].Nom,LesCategoriesDeBase[i].Couleur,True);
        end;
      //else if ((param1 = nil) and (param2 = nil)) then
        for i := 0 to NbCategoriecomp - 1 do
          begin
            if ConserverCat(LesCategoriesComp[i],Param1, Param2,SerieDonnees) then
            //if LesCategoriesComp[i].Active and (Not LesCategoriesComp[i].CatAnnexe) then
              begin
                if (param1 <> nil) and (param2 <> nil) then
                  EcritLeg(LesCategoriesComp[i].Nom,LesCategoriesComp[i].Couleur,True)
                else
                  EcritLeg(LesCategoriesComp[i].Nom,LesCategoriesComp[i].Couleur,False);
              end;
          end;
      if (param1 <> nil) and (param2 <> nil) then  // couleurs par densité
        begin
          X := 4+Round(Position.X);
          Y1 := 2+Round(Position.Y+NumLig*(htnm+2));
          Y2 := Y1 + Round(htnm)-1;
          LeBitmap.RenderText(X,Y1,'Densité -',qltt,clBlack32);
          X := X+3+LeBitmap.TextWidth('Densité -');
          LeBitmap.RenderText(X+21,Y1,'+',qltt,clBlack32);
          Y1 :=Y1+3;
          for i := 0 to 19 do
            LeBitmap.VertLineS(X+i,Y1,Y2,HSLtoRGB( 0.7 *(1 - i/19), 1, 0.5));
        end;
      LeBitmap.EndUpdate;
    end;
end;
procedure TEnsembleCategories.ChercheDependantes(LaCategorieP : TCategorie; Var CatDep : TGroupeCat);
Var
  i,j,n :integer;
begin
  n := 0;
  SetLength(CatDep,n);
  for i := 0 to NbCategorieComp - 1 do
    begin
      for j := 0 to Length(LesCategoriesComp[i].CatAscendants)- 1 do
        if LesCategoriesComp[i].CatAscendants[j] = LaCategorieP then
          begin
            inc(n);
            SetLength(CatDep,n);
            CatDep[n-1] := LesCategoriesComp[i];
            Break;
          end;
    end;
end;
procedure TEnsembleCategories.RAZResultats(Conserver : Boolean);//si Conserver conserve beNonDefini, sinon remet tout à beNonTeste
Var
  i : integer;
begin
  for i := 0 to NbCategorieDeBase - 1 do
    begin
      if Not(Conserver and (LesCategoriesDeBase[i].Resultat = beNonDefini)) then
        LesCategoriesDeBase[i].Resultat := beNonTeste;
    end;
  for i := 0 to NbCategorieComp - 1 do
    begin
      if Not(Conserver and (LesCategoriesComp[i].Resultat = beNonDefini)) then
        LesCategoriesComp[i].Resultat := beNonTeste;
    end;
end;
procedure TEnsembleCategories.AjouteDeBase(LaCategorie : TCategorie);
begin
  inc(NbCategorieDeBase);
  SetLength(LesCategoriesDeBase,NbCategorieDeBase);
  LesCategoriesDeBase[NbCategorieDeBase - 1] := LaCategorie;
  LaCategorie.Numero := NbCategorieDeBase - 1;
end;
procedure TEnsembleCategories.AjouteComp(LaCategorie : TCategorie);
begin
  inc(NbCategorieComp);
  SetLength(LesCategoriesComp,NbCategorieComp);
  LesCategoriesComp[NbCategorieComp - 1] := LaCategorie;
  LaCategorie.Numero := NbCategorieComp - 1;
end;

function TEnsembleCategories.LaCategorie (LaCouleur : TColor32) : TCategorie;
Var
  i : integer;
begin
  Result := nil;
  for i := NbCategorieComp -1 downto 0 do
    begin
      if LesCategoriesComp[i].Couleur = LaCouleur then
        begin
          Result := LesCategoriesComp[i];
          Break;
        end;
    end;
  if Result = nil then
    begin
      for i := NbCategorieDeBase -1 downto 0 do
        begin
          if LesCategoriesDeBase[i].Couleur = LaCouleur then
            begin
              Result := LesCategoriesDeBase[i];
              Break;
            end;
        end;
    end;
end;
function TEnsembleCategories.LeNumero(UneCategorie : TCategorie) : integer;
Var
  i : integer;
begin
  Result := -1;
  if UneCategorie <> nil then
    begin
      if UneCategorie.Composite then
        Result := -2-UneCategorie.Numero
      else
        Result := UneCategorie.Numero;
    end;
end;

//  *** TCategorie ***
Constructor TCategorie.Create;
begin
  inherited Create;
  AncienNom :='';
  CatAnnexe:=False;
  Resultat := beNonTeste;
  PolygoneDef := TPolygon32.Create;
end;

Destructor TCategorie.Destroy;
begin
  PolygoneDef.Free;
  Inherited Destroy;
end;

function TCategorie.EstAscendantDe (Cat : TCategorie) : Boolean;
Var
  i : integer;
begin
  Result := False;
  for i := 0 to Length(Cat.CatAscendants)-1 do
    if Cat.CatAscendants[i] = Self then
      begin
        Result := True;
        Break;
      end;
end;

procedure TCategorie.SetComposite (value : Boolean);
begin
  FComposite := value;
  if value then
    begin
      nbParams := 0;
      SetLength(Params,nbParams);
      SetLength(CatParents,2);
    end
  else
    SetLength(CatParents,0);
end;

procedure TCategorie.SetNbParams (nombre : integer);
begin
  if (nombre > 0) and (nombre < 3) then
    begin
      nbParams := nombre;
      SetLength(Params,nbParams);
    end;
end;
function TCategorie.Inclut(LaDonnee : Array of single;LesParametres : Array of TParam;ResolutionD : integer) :  TBoolEtendu;
  //(beNonDefini,beNonTeste,beTrue,beFalse);
Var
  i,j : integer;
  p : array of integer;
  Continuer : Boolean;
  Rep : array of Boolean;
  Rept : Boolean;
  Res : TBoolEtendu;
  Message,MesComp : string;
begin
  if Resultat <> beNonTeste then
    begin
{     if (Nom = 'Lymphocytes') and (Resultat = beFalse) then
      begin
        Message := Nom +';  - Ascendants : '+chr(10)+chr(13);
        for i := 0 to Length(CatAscendants)-1 do
          begin
            Message :=Message + InttoStr(i+1)+' '+ CatAscendants[i].Nom+' Resultat ';
            if CatAscendants[i].Resultat = beFalse then
              MesComp := 'Faux'
            else if CatAscendants[i].Resultat = beTrue then
              MesComp := 'Vrai'
            else if CatAscendants[i].Resultat = beNonTeste then
              MesComp := 'Non Teste'
            else if CatAscendants[i].Resultat = beNonDefini then
              MesComp := 'Non Défini';
            Message :=Message +MesComp+chr(10)+chr(13);
          end;
        MessageDlg(Message,mtWarning,[mbOk],0);
      end; }
      Result := Resultat;
      Exit;
    end;
  if Composite then
    begin
      for i := 0 to Length(CatAscendants)-1 do
        begin
          Res := CatAscendants[i].Resultat;
          if Res = beNonTeste then
            Res := CatAscendants[i].Inclut(LaDonnee,LesParametres,ResolutionD);
          if (Lien = tlEt) and (Res <> beTrue) then
            begin
              Resultat := Res;
              Result := Res;
              exit;
            end;
        end;
      SetLength(Rep,2);

      for i := 0 to 1 do
        begin
          if CatParents[i].Resultat = beNonDefini then
            begin
{              Message := CatParents[i].Nom +'; Non Defini - ';
              for j := 0 to Length(LesParametres)-1 do
                Message :=Message + 'Par'+InttoStr(i+1)+' '+ LesParametres[i].Nom+' ; ';
              for j := 0 to CatParents[i].nbParams - 1 do
                Message := Message + 'MonPar'+InttoStr(i+1)+' '+ CatParents[i].Params[i].Nom+' ; ';
              MessageDlg(Message,mtWarning,[mbOk],0);
}
              Resultat := beNonDefini;
              Result := Resultat;
              exit;
            end
          else if CatParents[i].Resultat = beTrue then
            Rep[i] := True
          else
            Rep[i] := False;
        end;
      Case Lien of  //(tlEt,TlOu,tlXor);
        tlEt : Rept := Rep[0] and Rep[1];
        tlOu : Rept := Rep[0] or Rep[1];
        tlXOr : Rept := Rep[0] Xor Rep[1];
      End;
      if Rept then
        Result := beTrue
      else
        Result := beFalse;
    end
  else
    begin
      //Tester si la donnée comporte les paramètres critères
      SetLength(p,nbParams);
      Continuer := True;
      for i := 0 to nbParams - 1 do
        begin
          p[i] := -1;
          for j := 0 to Length(LesParametres) - 1 do
            begin
              if Params[i] = LesParametres[j] then
                begin
                  p[i] := j;
                  Break;
                end;
            end;
          if p[i] = -1 then
            Continuer := False;
        end;
      if Continuer then
        begin
          Result := beFalse;
          if nbParams = 1 then
            begin
              if TestePoint(FloatPoint(LaDonnee[p[0]],0),ResolutionD) then
                Result := beTrue;
            end
          else if nbParams = 2 then
            if TestePoint(FloatPoint(LaDonnee[p[0]],LaDonnee[p[1]]),ResolutionD) then
              Result := beTrue;
        end
      else
        Result := beNonDefini;
    end;
  Resultat := Result;
end;
function TCategorie.TestePoint(LePoint: TFloatPoint;ResolutionD :integer):Boolean;
Var
  X,Y : Real;
begin
  Result := False;
  X := LePoint.X/ResolutionD*Resolution;
  if NbParams = 1 then
    Result := (X >= ValeurMin) and (X <= ValeurMax)
  else if NbParams = 2 then
    begin
      //Y := LePoint.Y/ResolutionD*Resolution;
      X :=X/DP16;
      Y := LePoint.Y/ResolutionD*Resolution/DP16;
      if Length(PolygoneDef.Points) > 0 then
        Result := PtInPolygon(Fixedpoint(X,Y),PolygoneDef.Points[0]);
    end;
  if Result then
    Resultat := beTrue
  else
    Resultat := beFalse;
end;
// *** TDonneesCyto ***
function TDonneesCyto.AjouteStreamFCS(CatRestr : TCategorie;ProgressBar:TProgressBar;Dernier : Boolean;Var MonFichier : TFileStream): Boolean;
Var
  OnePercent : integer;
  ENT1 : AnsiString;
  HeaderText : Ansistring;
  FinHeader : AnsiString;
  FSep : Char;
  BF : Array of Char;
  Buffer : PChar;
  NbParamSel : integer;
  i, j, np, nz : integer;
  ChP, ChE : string;
  ValeurEntiere : DWord;
  HeaderTextLast     :  INTEGER;   // Bytes 18-25
  HeaderDataFirst    :  INTEGER;   // Bytes 26-33
  HeaderDataLast     :  INTEGER;   // Bytes 34-41
  DataElementSize    :  INTEGER;   //
  NextData : int64;
  Vali : Byte;
  Conserver : Array of Boolean;
  NbConserves : integer;
  OffSetGen : int64;
  function StrNb(Nb : integer): string;
  Var
    i1, n : integer;
  begin
    Result := IntToStr(Nb);
    n := 8-length(Result);
    if n < 0 then
      Result :='       0'
    else
      begin
        for i1 := 1 to n do
          Result := ' '+Result;
      end;
  end;
begin
  Result := True;
  ProgressBar.Position := 0;
  ProgressBar.Visible := True;
  Application.ProcessMessages;
  OnePercent := NbEvenements div 100;

  NbParamSel := 0;
  for i := 0 to NbParam -1 do
    if Parametres[i].Sauvegarder then
      inc(NbParamSel);

  SetLength(Conserver,NbEvenements);
  if CatRestr = nil then
    begin
     NbConserves :=NbEvenements;
     for i := 0 to NbEvenements -1 do
       Conserver[i] := True;
    end
  else
    begin
     NbConserves :=0;
     LesCategories.RAZResultats(False);
     for i := 0 to NbEvenements -1 do
       begin
         LesCategories.RAZResultats(True);
         Conserver[i] := (CatRestr.Inclut(Donnees[i],Parametres,Resolution)= beTrue);
         if Conserver[i] then
           inc(NbConserves);
       end;
    end;

  if TypeDonnees = 'F' then
    DataElementSize := 4
  else
    DataElementSize := Parametres[0].Nombrebits div 8;


  FSep := chr(13);
  // le plus gros en premier pour évaluer la place nécessaire
  FinHeader :='$FIL'+FSep+UTF8ToAnsi(Nom)+FSep+'$TOT'+FSep+IntToStr(NbConserves)+FSep
              +'$PAR'+FSep+IntToStr(NbParamSel)+FSep+'$MODE'+FSep+'L'+FSep
              +'$BYTEORD'+FSep+'1,2,3,4'+FSep+'$DATATYPE'+FSep+TypeDonnees+FSep+'$DATE'+FSep+LaDate+FSep;
  if Cytometre <>'' then
    FinHeader := FinHeader+'$CYT'+FSep+Cytometre+FSep;
  if Projet <>'' then
    FinHeader := FinHeader+'$PROJ'+FSep+UTF8ToAnsi(Projet)+FSep;
  if Cellules <>'' then
    FinHeader := FinHeader+'$CELLS'+FSep+UTF8ToAnsi(Cellules)+FSep;
  if ParamSerie <>'' then
    begin
      FinHeader := FinHeader+'NOMPARAMSERIE'+FSep+UTF8ToAnsi(ParamSerie)+FSep;
      if UniteParamSerie <>'' then
        FinHeader := FinHeader+'UNITPARAMSERIE'+FSep+UTF8ToAnsi(UniteParamSerie)+FSep;
      FinHeader := FinHeader+'VALPARAMSERIE'+FSep+IntToStr(ValeurParamSerie)+FSep;
    end;

  np := -1;
  for i := 0 to NbParam -1 do
    begin
      if Parametres[i].Sauvegarder then
        begin
          inc(np);
          ChP := '$P'+IntToStr(np+1);
          ChE :=IntToStr(Parametres[i].Puis10Log)+','; //  E
          if (Parametres[i].Puis10Log = 0) or (Parametres[i].ValPourZero <= 0) then
            ChE := ChE+'0'
          else if Parametres[i].ValPourZero >= 1 then
            ChE := ChE+IntToStr(Round(Parametres[i].ValPourZero))
          else
            begin
              ChE := ChE+'0';
              nz := Round(-log10(Parametres[i].ValPourZero));
              if nz > 0 then
                begin
                  ChE := ChE+'.';
                  for j := 1 to nz-1 do
                    ChE := ChE+'0';
                  ChE := ChE+'1';
                end;
            end;
          FinHeader := FinHeader+ChP+'N'+FSep+UTF8ToAnsi(Parametres[i].Nom)+FSep
                       +ChP+'R'+FSep+IntToStr(Parametres[i].Resolution)+FSep
                       +ChP+'B'+FSep+IntToStr(Parametres[i].Nombrebits)+FSep
                       +ChP+'E'+FSep+ChE+FSep;
          if Parametres[i].Alias <> '' then
            FinHeader := FinHeader+ChP+'S'+FSep+UTF8ToAnsi(Parametres[i].Alias)+FSep;
        end;
    end;

  HeaderDataFirst := (((256+112+Length(FinHeader)) div 4)+3)*4;
  HeaderDataLast := HeaderDataFirst+NbParamSel*NbConserves*DataElementSize-1;
  if Dernier then
    NextData := 0
  else
    NextData := ((HeaderDataLast+2) div 4)*4 +4;

  FinHeader :=FinHeader+'$NEXTDATA'+FSep+IntToStr(NextData)+FSep;

  // puis le début
  HeaderText :=FSep;     //Séparateur en premier
  HeaderText := HeaderText + '$BEGINANALYSIS'+FSep+'0'+FSep+'$ENDANALYSIS'+FSep+'0'+FSep
                        + '$BEGINSTEXT'+FSep+'0'+FSep+'$ENDSTEXT'+FSep+'0'+FSep
                        + '$BEGINDATA'+FSep+IntToStr(HeaderDataFirst)+FSep+'$ENDDATA'+FSep+IntToStr(HeaderDataLast)+FSep;

  HeaderText := HeaderText + FinHeader;
  HeaderTextLast := 256 + Length(HeaderText)-1;

  ENT1 := 'FCS3.0         256'+ StrNb(HeaderTextLast)+ StrNb(HeaderDataFirst)+ StrNb(HeaderDataLast)
          +'       0       0';
  For i := Length(ENT1) to 256 do
    ENT1 := ENT1+' ';
  OffSetGen := MonFichier.Position;
  for i := 1 to 256 do
    begin
      Vali := ord(ENT1[i]);
      MonFichier.WriteBuffer(Vali,1);
    end;
  for i := 1 to Length(HeaderText) do
    begin
      Vali := ord(HeaderText[i]);
      MonFichier.WriteBuffer(Vali,1);
    end;
 Vali := 32;
 For i:= MonFichier.Position to OffsetGen+HeaderDataFirst -1 do
   MonFichier.WriteBuffer(Vali,1);

 //  Enregistrement des données
  for i := 0 to NbEvenements -1 do  //** si CatRestr <>ni implanter la restriction à une categorie
    begin
      if i mod OnePercent = 0 then
        begin
          ProgressBar.Position := Round(100 * i / NbEvenements);
          Application.ProcessMessages;
        end;
      if Conserver[i] then
        begin
	  for j := 0 to NbParam -1 do
	    if Parametres[j].Sauvegarder then
	      begin
	        if TypeDonnees = 'F' then
	          MonFichier.Write(Donnees[i,j],DataElementSize)
	        else
	          begin
	            ValeurEntiere := Round(Donnees[i,j]);
	            MonFichier.Write(ValeurEntiere,DataElementSize);
	          end;
	      end;
        end;
    end;
  // Ajouter les blancs s'il y a un autre jeu de données après
  if Not Dernier then
    for i := 1 to  (NextData-HeaderDataLast-1) do
      begin
        Vali := $20;
        MonFichier.WriteBuffer(Vali,1);
      end;
end;

function TDonneesCyto.enregistreFCS(NouvNomFichier: string;CatRestr : TCategorie;ProgressBar:TProgressBar) : Boolean;
Var
  MonFichier : TFileStream;
  OnePercent : integer;
  NmF : Ansistring;
  ENT1 : AnsiString;
  HeaderText : Ansistring;
  FinHeader : AnsiString;
  FSep : Char;
  BF : Array of Char;
  Buffer : PChar;
  NbParamSel : integer;
  i, j, np, nz : integer;
  ChP, ChE : string;
  ValeurEntiere : DWord;
  HeaderTextLast     :  INTEGER;   // Bytes 18-25
  HeaderDataFirst    :  INTEGER;   // Bytes 26-33
  HeaderDataLast     :  INTEGER;   // Bytes 34-41
  DataElementSize    :  INTEGER;   // 2 for 'I' or 4 for 'F' DATATYPE
  Vali : Byte;
  Conserver : Array of Boolean;
  NbConserves : integer;
  function StrNb(Nb : integer): string;
  Var
    i1, n : integer;
  begin
    Result := IntToStr(Nb);
    n := 8-length(Result);
    if n < 0 then
      Result :='       0'
    else
      begin
        for i1 := 1 to n do
          Result := ' '+Result;
      end;
  end;
begin
  Result := False;
  ProgressBar.Position := 0;
  ProgressBar.Visible := True;
  Application.ProcessMessages;
  OnePercent := NbEvenements div 100;

  NbParamSel := 0;
  for i := 0 to NbParam -1 do
    if Parametres[i].Sauvegarder then
      inc(NbParamSel);

  SetLength(Conserver,NbEvenements);
  if CatRestr = nil then
    begin
     NbConserves :=NbEvenements;
     for i := 0 to NbEvenements -1 do
       Conserver[i] := True;
    end
  else
    begin
     NbConserves :=0;
     LesCategories.RAZResultats(False);
     for i := 0 to NbEvenements -1 do
       begin
         LesCategories.RAZResultats(True);
         Conserver[i] := (CatRestr.Inclut(Donnees[i],Parametres,Resolution)= beTrue);
         if Conserver[i] then
           inc(NbConserves);
       end;
    end;

  if TypeDonnees = 'F' then
    DataElementSize := 4
  else
    DataElementSize := Parametres[0].Nombrebits div 8;

  NmF := UTF8ToAnsi(NouvNomFichier);

  FSep := chr(13);

  // le plus gros en premier pour évaluer la place nécessaire
  FinHeader :='$FIL'+FSep+UTF8ToAnsi(ExtractFileName(NouvNomFichier))+FSep+'$TOT'+FSep+IntToStr(NbConserves)+FSep
              +'$PAR'+FSep+IntToStr(NbParamSel)+FSep+'$MODE'+FSep+'L'+FSep
              +'$BYTEORD'+FSep+'1,2,3,4'+FSep+'$DATATYPE'+FSep+TypeDonnees+FSep
              +'$NEXTDATA'+FSep+'0'+FSep+'$DATE'+FSep+LaDate+FSep;
  if Cytometre <>'' then
    FinHeader := FinHeader+'$CYT'+FSep+Cytometre+FSep;
  if Projet <>'' then
    FinHeader := FinHeader+'$PROJ'+FSep+UTF8ToAnsi(Projet)+FSep;
  if Cellules <>'' then
    FinHeader := FinHeader+'$CELLS'+FSep+UTF8ToAnsi(Cellules)+FSep;
  np := -1;
  for i := 0 to NbParam -1 do
    begin
      if Parametres[i].Sauvegarder then
        begin
          inc(np);
          ChP := '$P'+IntToStr(np+1);
          ChE :=IntToStr(Parametres[i].Puis10Log)+','; //  E
          if (Parametres[i].Puis10Log = 0) or (Parametres[i].ValPourZero <= 0) then
            ChE := ChE+'0'
          else if Parametres[i].ValPourZero >= 1 then
            ChE := ChE+IntToStr(Round(Parametres[i].ValPourZero))
          else
            begin
              ChE := ChE+'0';
              nz := Round(-log10(Parametres[i].ValPourZero));
              if nz > 0 then
                begin
                  ChE := ChE+'.';
                  for j := 1 to nz-1 do
                    ChE := ChE+'0';
                  ChE := ChE+'1';
                end;
            end;
          FinHeader := FinHeader+ChP+'N'+FSep+UTF8ToAnsi(Parametres[i].Nom)+FSep
                       +ChP+'R'+FSep+IntToStr(Parametres[i].Resolution)+FSep
                       +ChP+'B'+FSep+IntToStr(Parametres[i].Nombrebits)+FSep
                       +ChP+'E'+FSep+ChE+FSep;
          if Parametres[i].Alias <> '' then
            FinHeader := FinHeader+ChP+'S'+FSep+UTF8ToAnsi(Parametres[i].Alias)+FSep;
        end;
    end;

  HeaderDataFirst := (((256+100+Length(FinHeader)) div 4)+3)*4;
  HeaderDataLast := HeaderDataFirst+NbParamSel*NbConserves*DataElementSize-1;
  // puis le début
  HeaderText :=FSep;     //Séparateur en premier
  HeaderText := HeaderText + '$BEGINANALYSIS'+FSep+'0'+FSep+'$ENDANALYSIS'+FSep+'0'+FSep
                        + '$BEGINSTEXT'+FSep+'0'+FSep+'$ENDSTEXT'+FSep+'0'+FSep
                        + '$BEGINDATA'+FSep+IntToStr(HeaderDataFirst)+FSep+'$ENDDATA'+FSep+IntToStr(HeaderDataLast)+FSep;

  HeaderText := HeaderText + FinHeader;
  HeaderTextLast := 256 + Length(HeaderText)-1;

  ENT1 := 'FCS3.0         256'+ StrNb(HeaderTextLast)+ StrNb(HeaderDataFirst)+ StrNb(HeaderDataLast)
          +'       0       0';
  For i := Length(ENT1) to 256 do
    ENT1 := ENT1+' ';

  MonFichier := TFileStream.Create(NmF,fmOpenWrite or fmCreate);
  try
    for i := 1 to 256 do
      begin
        Vali := ord(ENT1[i]);
        MonFichier.WriteBuffer(Vali,1);
      end;
    for i := 1 to Length(HeaderText) do
      begin
        Vali := ord(HeaderText[i]);
        MonFichier.WriteBuffer(Vali,1);
      end;
   Vali := 32;
   For i:= MonFichier.Position to HeaderDataFirst -1 do
     MonFichier.WriteBuffer(Vali,1);

   //  Enregistrement des données
    for i := 0 to NbEvenements -1 do  //** si CatRestr <>ni implanter la restriction à une categorie
      begin
        if i mod OnePercent = 0 then
          begin
            ProgressBar.Position := Round(100 * i / NbEvenements);
            Application.ProcessMessages;
          end;
        if Conserver[i] then
          begin
	    for j := 0 to NbParam -1 do
	      if Parametres[j].Sauvegarder then
	        begin
	          if TypeDonnees = 'F' then
	            MonFichier.Write(Donnees[i,j],DataElementSize)
	          else
	            begin
	              ValeurEntiere := Round(Donnees[i,j]);
	              MonFichier.Write(ValeurEntiere,DataElementSize);
	            end;
	        end;
          end;
      end;
  finally
    MonFichier.Free;
    ProgressBar.Visible := False;
    Application.ProcessMessages;
  end;
end;
function TDonneesCyto.NumParam(LeParam : TParam) : integer;
Var
  i : integer;
begin
  Result := -1;
  for i := 0 to NbParam-1 do
    if Parametres[i] = LeParam then
      begin
        Result := i;
        Break;
      end;
end;
function TDonneesCyto.NumParamTaille : integer;
Var
  i : integer;
  ChRec : string;
begin
  Result := -1;
  ChRec := 'TAILLE';
  for i := 0 to NbParam-1 do
    if UpperCase(Parametres[i].Alias) = ChRec then
      begin
        Result := i;
        Break;
      end;
  if Result = -1 then
    begin
      ChRec := 'FSC-A';
      for i := 0 to NbParam-1 do
      if UpperCase(Parametres[i].Nom) = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'FSC-H';
      for i := 0 to NbParam-1 do
      if UpperCase(Parametres[i].Nom) = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'FSC';
      for i := 0 to NbParam-1 do
      if Pos(ChRec,UpperCase(Parametres[i].Nom)) > 0  then
        begin
          Result := i;
          Break;
        end;
    end;
end;
function TDonneesCyto.NumParamComplexite : integer;
Var
  i : integer;
  ChRec : string;
begin
  Result := -1;
  ChRec := 'Granulosité';
  for i := 0 to NbParam-1 do
    if Parametres[i].Alias = ChRec then
      begin
        Result := i;
        Break;
      end;
  if Result = -1 then
    begin
      ChRec := 'Granularité';
      for i := 0 to NbParam-1 do
      if Parametres[i].Nom = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'Granulométrie';
      for i := 0 to NbParam-1 do
      if Parametres[i].Nom = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'GRANULOSITE';
      for i := 0 to NbParam-1 do
      if UpperCase(Parametres[i].Nom) = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'GRANULOMETRIE';
      for i := 0 to NbParam-1 do
      if UpperCase(Parametres[i].Nom) = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'GRANULARITE';
      for i := 0 to NbParam-1 do
      if UpperCase(Parametres[i].Nom) = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'SSC-A';
      for i := 0 to NbParam-1 do
      if UpperCase(Parametres[i].Nom) = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'SSC-H';
      for i := 0 to NbParam-1 do
      if UpperCase(Parametres[i].Nom) = ChRec then
        begin
          Result := i;
          Break;
        end;
    end;
  if Result = -1 then
    begin
      ChRec := 'SSC';
      for i := 0 to NbParam-1 do
      if Pos(ChRec,UpperCase(Parametres[i].Nom)) > 0  then
        begin
          Result := i;
          Break;
        end;
    end;
end;
procedure TDonneesCyto.ChercheCouleurs(CatRestr : TCategorie;CoulDens : Boolean;Param1,Param2  : TParam);
Var
  i,j,p : integer;
  LePoint : TFloatPoint;
  LaCouleur : TColor32;
  LaCategorie : TCategorie;
begin
  //Activation de toutes les catégories
  for i := 0 to LesCategories.NbCategorieDeBase -1 do
    LesCategories.LesCategoriesDeBase[i].Active:=True;
  if CoulDens then
    begin
      for i := 0 to LesCategories.NbCategorieComp -1 do
        LesCategories.LesCategoriesComp[i].Active:=False;
    end
  else
    begin
      for i := 0 to LesCategories.NbCategorieComp -1 do
        LesCategories.LesCategoriesComp[i].Active:=True;
    end;

  // Sauf CatRestr et ses ascendants
  if CatRestr <> nil then
    begin
      for i := 0 to Length(CatRestr.CatAscendants)-1 do
        CatRestr.CatAscendants[i].Active:=False;
      CatRestr.Active:=False;
    end;
  //** restreindre plus si couleurs par densité
  if CoulDens then
    begin
      for i := 0 to LesCategories.NbCategorieDeBase -1 do
        begin
          LaCategorie := LesCategories.LesCategoriesDeBase[i];
          LaCategorie.Active:=LaCategorie.Active and
            (((LaCategorie.Params[0] = Param1) and (LaCategorie.Params[1] = Param2))
             or ((LaCategorie.Params[0] = Param2) and (LaCategorie.Params[1] = Param1)));
          if LaCategorie.Active and (LaCategorie.CatFille<>nil) then
            begin
              LaCategorie.Active := False;
              LaCategorie.CatFille.Active:=True;
            end;
        end;
    end;
      LesCategories.RAZResultats(False);
      for i := 0 to NbEvenements - 1 do
        begin
         // Couleurs[i] := $FFFFFFFF;
         // if (CatRestr= nil) or (CatRestr.Inclut(Donnees[i],Parametres,Resolution)= beTrue) then
           // begin
              Couleurs[i] := $FF000000;
              LesCategories.RAZResultats(True);
              for j := LesCategories.NbCategorieDeBase -1 downto 0 do
                begin
                  LaCategorie := LesCategories.LesCategoriesDeBase[j];
                  if LaCategorie.Active  and (Not LaCategorie.CatAnnexe)  //and (LaCategorie<>CatRestr) Active a été mis à false
                      and (LaCategorie.Inclut(Donnees[i],Parametres,Resolution) = beTrue) then
                    begin
                      Couleurs[i] := LaCategorie.Couleur;
                      Break;
                    end;
                end;

              // On cherche ensuite dans les Catégories composées
              for j := LesCategories.NbCategorieComp -1 downto 0 do
                begin
                  LaCategorie := LesCategories.LesCategoriesComp[j];
                  if LaCategorie.Active  and (Not LaCategorie.CatAnnexe)  //and (LaCategorie<>CatRestr)
                      and (LaCategorie.Inclut(Donnees[i],Parametres,Resolution) = beTrue) then
                    begin
                      Couleurs[i] := LaCategorie.Couleur;
                      Break;
                    end;
                end;
              if (CatRestr <>nil) and Not(CatRestr.Inclut(Donnees[i],Parametres,Resolution)= beTrue) then
                Couleurs[i] := $FFFFFFFF;
            //end;
        end;
end;
procedure TDonneesCyto.CalculeStats;
Var
  i,j,k,p : integer;
  LePoint : TFloatPoint;
  LaCouleur : TColor32;
  LaCategorie : TCategorie;
  Vali : Array of real;
  Vali2 : Array of real;
begin
  if NbEvenements > 1 then
    begin
      LesCategories.RAZResultats(False);

      SetLength(MoysP,NbParam);
      SetLength(EcTsP,NbParam);
      SetLength(NbEventCS,LesCategories.NbCategorieDeBase);
      SetLength(MoysPCS,LesCategories.NbCategorieDeBase);
      SetLength(EcTsPCS,LesCategories.NbCategorieDeBase);
      for i := 0 to NbParam -1 do
        begin
          MoysP[i] := 0;
          EcTsP[i] := 0;
        end;
      for i := 0 to LesCategories.NbCategorieDeBase - 1 do
        begin
          SetLength(MoysPCS[i],NbParam);
          SetLength(EcTsPCS[i],NbParam);
          NbEventCS[i] := 0;
          for j := 0 to NbParam -1 do
            begin
              MoysPCS[i,j] := 0;
              EcTsPCS[i,j] := 0;
            end;
        end;
      SetLength(NbEventCC,LesCategories.NbCategorieComp);
      SetLength(MoysPCC,LesCategories.NbCategorieComp);
      SetLength(EcTsPCC,LesCategories.NbCategorieComp);

      for i := 0 to LesCategories.NbCategorieComp - 1 do
        begin
          SetLength(MoysPCC[i],NbParam);
          SetLength(EcTsPCC[i],NbParam);
          NbEventCC[i] := 0;
          for j := 0 to NbParam -1 do
            begin
              MoysPCC[i,j] := 0;
              EcTsPCC[i,j] := 0;
            end;
        end;
      SetLength(Vali,NbParam);
      SetLength(Vali2,NbParam);
      for i := 0 to NbEvenements - 1 do
        begin
          for j := 0 to NbParam -1 do
            begin
              if Parametres[j].Puis10Log = 0 then
                Vali[j] :=Donnees[i,j]
              else
                Vali[j] := power(10,Donnees[i,j]/Parametres[j].Resolution*Parametres[j].Puis10Log);
              Vali2[j] := Vali[j] * Vali[j];
              MoysP[j] := MoysP[j] + Vali[j];
              EcTsP[j] := EcTsP[j] + Vali2[j];
            end;
          Couleurs[i] := $FF000000;
          LesCategories.RAZResultats(True);
          for j := 0 to LesCategories.NbCategorieDeBase -1 do
            begin
              LaCategorie := LesCategories.LesCategoriesDeBase[j];
              //if LaCategorie.Active and (LaCategorie.Inclut(Donnees[i],Parametres,Resolution) = beTrue) then
              if (LaCategorie.Inclut(Donnees[i],Parametres,Resolution) = beTrue) then
                begin
                  inc(NbEventCS[j]);
                  for k := 0 to NbParam -1 do  // Sommes et sommes des carrés
                    begin
                      MoysPCS[j,k] := MoysPCS[j,k] + Vali[k];
                      EcTsPCS[j,k] := EcTsPCS[j,k] + Vali2[k];
                    end;
                end;
            end;

          // On cherche ensuite dans les Catégories composées
          for j := 0 to LesCategories.NbCategorieComp - 1 do
            begin
              LaCategorie := LesCategories.LesCategoriesComp[j];
              //if LaCategorie.Active and (LaCategorie.Inclut(Donnees[i],Parametres,Resolution) = beTrue) then
              if (LaCategorie.Inclut(Donnees[i],Parametres,Resolution) = beTrue) then
                begin
                  inc(NbEventCC[j]);
                  for k := 0 to NbParam -1 do  // Sommes et sommes des carrés
                    begin
                      MoysPCC[j,k] := MoysPCC[j,k] + Vali[k];
                      EcTsPCC[j,k] := EcTsPCC[j,k] + Vali2[k];
                    end;
                end;
            end;
        end;
      // calculs complémentaires

      for i := 0 to NbParam -1 do
        begin
          MoysP[i] := MoysP[i] / NbEvenements;
          EcTsP[i] := Sqrt((EcTsP[i] / NbEvenements) - (MoysP[i] * MoysP[i]));
        end;
      for i := 0 to LesCategories.NbCategorieDeBase - 1 do
        begin
          for j := 0 to NbParam -1 do
            begin
              if NbEventCS[i] > 1 then
                begin
                  MoysPCS[i,j] := MoysPCS[i,j] / NbEventCS[i];
                  EcTsPCS[i,j] := Sqrt((EcTsPCS[i,j]/NbEventCS[i]) - (MoysPCS[i,j]*MoysPCS[i,j]));
                end
              else
                EcTsPCS[i,j] := 0;
            end;
        end;

      for i := 0 to LesCategories.NbCategorieComp - 1 do
        begin
          for j := 0 to NbParam -1 do
            begin
              if NbEventCC[i] > 1 then
                begin
                  MoysPCC[i,j] := MoysPCC[i,j] / NbEventCC[i];
                  EcTsPCC[i,j] := Sqrt((EcTsPCC[i,j]/NbEventCC[i]) - (MoysPCC[i,j]*MoysPCC[i,j]));
                end
              else
                EcTsPCC[i,j] := 0;
            end;
        end;
    end;
end;

function ConversionLog(Valeur : real):real;
begin
  Result := 0;
  if valeur >= 10 then
    Result := 200*Log10(Valeur/10);
end;

function ConversionExp(Valeur : real) : real;
begin
  Result := 10*Power(10,Valeur/200);
end;
function FracAleatoire(Valeur : Real) : real;
begin
  Result := Valeur;
  if Valeur > 1 then
    Result := int(Valeur) + Random;
end;

procedure TDonneesCyto.TraceHisto(LeBitmap : TBitmap32 ; LeParam : TParam;ValXEchLog, Corr : Boolean;CatRestr : TCategorie;Var MaxiX,MaxiY : real);
Var
  i : integer;
  dx,dy :real;
  X,XF,Y,Yb : integer;
  H : Array of integer;
  Hmax : integer;
  lgt,htt :real;
  MaCategorie : TCategorie;
  MonNumParam: integer;
  Inverser : Boolean;
  CouleurL : TColor32;
  LesGraduations : TGraduations;
  NbClasses,divis: integer;
  Catmax : integer;
  TitreX : string;
  XT : integer;
  Valeur : real;
  MiniX,MiniY : integer;
  RangeCondition : Boolean;
  Xr : real;
  decade : real;
  larg : integer;
  NbGX : integer;
  Vx : real;
  divisr : real;
  iclasse : integer;
begin
  if UpperCase(LeParam.Nom) = 'TIME' then
    begin
      MessageDlg('Le temps ne se prête pas au tracé d''un histogramme',mtError,[mbOK],0);
    end
  else
    begin
       //Activation de toutes les catégories
      for i := 0 to LesCategories.NbCategorieDeBase -1 do
        LesCategories.LesCategoriesDeBase[i].Active:=True;
      for i := 0 to LesCategories.NbCategorieComp -1 do
        LesCategories.LesCategoriesComp[i].Active:=True;
      // Sauf CatRestr et ses ascendants
      if CatRestr <> nil then
        begin
          for i := 0 to Length(CatRestr.CatAscendants)-1 do
            CatRestr.CatAscendants[i].Active:=False;
          CatRestr.Active:=False;
        end;
      MonNumParam := NumParam(LeParam);
      LeBitmap.BeginUpdate;
      if ValXEchLog then
        begin
          MaxiX := 200*log10(LeParam.Resolution);
          MiniX := 200;
          if MaxiX <= 2048 then
            NbClasses := Round(MaxiX)
          else
            NbClasses := 1024;
          Randomize;
        end
      else
        begin
          if LeParam.TypeEchelle = TeLog then
            MaxiX := LeParam.Resolution
          else
            begin
	      if Maxi2[MonNumParam] = 0 then
	         MaxiX := Maxi[MonNumParam]
	      else
	        MaxiX := Maxi2[MonNumParam];
            end;
          MiniX := 1;
	  if MaxiX <= 2048 then
	    NbClasses := Round(MaxiX)
	  else
	    NbClasses := 1024;
        end;
      divis := Round(MaxiX) div NbClasses;
      if ValXEchLog then
        divisr := MaxiX/(NbClasses-1);
      dx := (LeBitmap.Width-MargeG-MargeD)/NbClasses;
      // Tracé des axes
      LeBitmap.VertLineS(MargeG,MargeH,LeBitmap.Height-MargeB,clBlack32);
      LeBitmap.HorzLineS(MargeG,LeBitmap.Height-MargeB,LeBitmap.Width-MargeD,clBlack32);
      SetLength(H,NbClasses);
      For i :=0 to NbClasses -1 do
        H[i] := 0;
      Hmax := 0;
      for i := 0 to NbEvenements - 1 do
        begin
          if ValXEchLog then
            RangeCondition := InRange(Donnees[i,MonNumParam],10, LeParam.Resolution -2)
          else
            RangeCondition := InRange(Donnees[i,MonNumParam],1, LeParam.Resolution -2);
          LesCategories.RAZResultats(True);
          RangeCondition := RangeCondition and ((CatRestr= nil) or (CatRestr.Inclut(Donnees[i],Parametres,Resolution)= beTrue));
	      if RangeCondition then
	        begin
	          if ValXEchLog then
                    begin
		      if Corr then
                        Valeur := 200*(log10(FracAleatoire(Donnees[i,MonNumParam]))-1)
                      else
       	                Valeur := 200*(log10(Donnees[i,MonNumParam])-1);
                      iclasse := Round(Valeur/divisr);
	              if (iclasse < NbClasses) then
	                inc(H[iclasse]);
                    end
              else
                begin
		  Valeur := Donnees[i,MonNumParam];
	          if ((Round(Valeur)div divis)< NbClasses) then
	            inc(H[Round(Valeur)div divis]);
                end;
            end;
        end;
      Hmax := MaxIntValue(H);
      if Hmax > 0 then
        begin
          if (Hmax = H[0]) and (H[1]>0)and((Hmax/H[1]) > 5) then
            HMax := H[1];
          Catmax := NbClasses -1;
          if LeParam.TypeEchelle = TeLin then
            begin
              for i := NbClasses -1 downto 0 do
                begin
                  if H[i] < Hmax/1000 then
                    Catmax := i
                  else
                    Break;
                end;
              if Catmax < NbClasses*0.7 then
                begin
                  MaxiX := Catmax * divis;
                  divis := Round(MaxiX) div NbClasses;
                  if divis = 0 then
                    begin
                      NbClasses := Round(MaxiX);
                      divis := 1;
                    end;
                  if ValXEchLog then
                    divisr := MaxiX/(NbClasses);  // - 1
                  SetLength(H,NbClasses);

                  For i :=0 to NbClasses -1 do
                    H[i] := 0;
                  Hmax := 0;
                  for i := 0 to NbEvenements - 1 do
                    begin
	              if ValXEchLog then
	                RangeCondition := InRange(Donnees[i,MonNumParam],10, LeParam.Resolution -2)
	              else
	                RangeCondition := InRange(Donnees[i,MonNumParam],1, LeParam.Resolution -2);
                      LesCategories.RAZResultats(True);
                      RangeCondition := RangeCondition and ((CatRestr= nil) or (CatRestr.Inclut(Donnees[i],Parametres,Resolution)= beTrue));
	              if RangeCondition then
	                begin
                          if ValXEchLog then
	                     begin
		               if Corr then
                                 Valeur := 200*(log10(FracAleatoire(Donnees[i,MonNumParam]))-1)
		               else
                                 Valeur := 200*(log10(Donnees[i,MonNumParam])-1);
                               iclasse := Round(Valeur/divisr);
    	                       if (iclasse < NbClasses) then
    	                         inc(H[iclasse]);
	                      end
                          else
                            begin
	    	              Valeur := Donnees[i,MonNumParam];
	                        if ((Round(Valeur)div divis)< NbClasses) then
	                          inc(H[Round(Valeur)div divis]);
                            end;
	                end;
                    end;
                  Hmax := MaxIntValue(H);
                  if (Hmax = H[0]) and ((Hmax/H[1]) > 5) then
                    HMax := H[1];
                end;
            end;
          dx := (LeBitmap.Width-MargeG-MargeD)/NbClasses;
          if MaxiY = 0 then
            MaxiY := HMax;
          LesGraduations := CalculeGraduations(LeBitmap,MaxiX,MaxiY,LeParam,nil,ValXEchLog,False);
          for i := 0 to Length(LesGraduations.PX)-1 do
            begin
              LeBitmap.VertLineS(Round(LesGraduations.PX[i]),LeBitmap.Height-MargeB,LeBitmap.Height-MargeB+5,clBlack32);
              lgt := LeBitmap.TextWidth(LesGraduations.VX[i])/2;
              htt := LeBitmap.TextHeight(LesGraduations.VX[i]);
              XT := Round(LesGraduations.PX[i]-lgt);
              if (XT + 2*lgt) > (LeBitmap.Width - 2) then
                XT := LeBitmap.Width - 2 - Round(2*lgt);
              LeBitmap.RenderText(XT,LeBitmap.Height-MargeB+6,LesGraduations.VX[i],qltt,clBlack32);
            end;
          for i := 0 to Length(LesGraduations.PY)-1 do
            begin
              LeBitmap.HorzLineS(margeG,LesGraduations.PY[i],margeG-5,clBlack32);
              lgt := LeBitmap.TextWidth(LesGraduations.VY[i]);
              htt := LeBitmap.TextHeight(LesGraduations.VY[i])/2;
              LeBitmap.RenderText(Round(margeG-6-lgt),Round(LesGraduations.PY[i]-htt),LesGraduations.VY[i],qltt,clBlack32);
            end;
          LesGraduations.Free;
          // Titres des axes
          if LeParam.Alias = '' then
            TitreX := LeParam.Nom
          else
            TitreX := LeParam.Alias;
          lgt :=LeBitmap.TextWidth(TitreX);
          htt := LeBitmap.TextHeight(TiTreX);
          Yb := LeBitmap.Height-MargeB;
          LeBitmap.RenderText(Round(MargeG+(LeBitmap.Width - lgt)/2),LeBitmap.Height-Round(htt)-6,TitreX,qltt,clBlack32);
          LeBitmap.RenderText(2,2,'Nombre',qltt,clBlack32);

          dy := (LeBitmap.Height-MargeB-MargeH)/MaxiY;
          For i :=0 to  NbClasses-1 do
            begin
              if ValXEchLog then  // 200*(log10(Donnees[i,MonNumParam])-1) //Round(Valeur)div divis
                begin
                  if i < NbClasses - 1 then
                    Vx :=10*Power(10,(i*divisr)/200)
                  else
                    Vx :=10*Power(10,(i*divisr)/200);
                  MaCategorie := LaCategorieHisto(VX,LeParam,LeParam.Resolution);
                end
              else
                MaCategorie := LaCategorieHisto(i*divis,LeParam,LeParam.Resolution);
              if MaCategorie <> nil then
                begin
                  if MaCategorie.CatAnnexe and (MaCategorie.CatFille <> nil) then
                    CouleurL := MaCategorie.CatFille.Couleur
                  else
                    CouleurL := MaCategorie.Couleur;
                end
              else
                CouleurL := $FF000000;
              {if Vx < 100 then
                CouleurL := clgreen32
              else if Vx < 1000 then
                CouleurL := clfuchsia32
              else if Vx < 10000 then
                CouleurL := clred32
              else
                CouleurL := clblue32;
              }
              X := Round(MargeG+i*dX);
              XF := Round(MargeG+(i+1)*dX);
              if XF = X then
                XF := X+1;
              Y := Round(MargeH + (MaxiY - H[i])*dY);
              LeBitmap.FillRectS(Rect(X,Y,XF,Yb),CouleurL);
            end;
        end;
      LeBitmap.EndUpdate;
    end;
end;
procedure TDonneesCyto.TraceGraphe(LeBitmap : TBitmap32 ; Param1,Param2  : TParam ; Taillepoints : integer;ValXEchLog,ValYEchLog : Boolean;CatRestr : TCategorie;CoulDens : Boolean;Var MaxiX,MaxiY : real);
Var
  i : integer;
  dx,dy :real;
  X,XF,Y,YF,XT : integer;
  LesGraduations : TGraduations;
  lgt,htt : real;
  Numparam1,Numparam2  : integer;
  //Inverser : Boolean;
  MaCategorie : TCategorie;
  MonPolyg : TPolygon32;
  TitreX,TitreY :string;
  ValX,ValY : real;
  largeur, hauteur : integer;
  CouleurP : TColor32;
begin
  Numparam1 := Numparam(Param1);
  Numparam2 := Numparam(Param2);
  TitreX := Param1.Alias;
  if TitreX ='' then
    TitreX := Param1.Nom;
  TitreY := Param2.Alias;
  if TitreY ='' then
    TitreY := Param2.Nom;
  ChercheCouleurs(CatRestr,CoulDens,Param1,Param2);
  LeBitmap.BeginUpdate;
  if ValXEchLog then
    begin
      MaxiX := 200*log10(Maxi[NumParam1]);
      Randomize;
    end
  else
    begin
      if Param1.TypeEchelle = teLog then
        MaxiX := Param1.Resolution
      else
        begin
          if Maxi2[NumParam1] > 0 then
            MaxiX := Maxi2[NumParam1]
          else
            MaxiX := Maxi[NumParam1];
        end;
    end;
  if ValYEchLog then
    begin
      MaxiY := 200*log10(Maxi[NumParam2]);
      Randomize;
    end
  else
    begin
      if Param2.TypeEchelle = teLog then
        MaxiY := Param2.Resolution
      else
        begin
          if Maxi2[NumParam2] > 0 then
            MaxiY := Maxi2[NumParam2]
          else
            MaxiY := Maxi[NumParam2];
      end;
    end;
  largeur := LeBitmap.Width-MargeG-MargeD;
  hauteur := LeBitmap.Height-MargeH-MargeB;
  dx := largeur/MaxiX;  //Resolution
  dy := hauteur/MaxiY; //Resolution
  // Tracé des axes
  LeBitmap.VertLineS(MargeG,MargeH,LeBitmap.Height-MargeB,clBlack32);
  LeBitmap.HorzLineS(MargeG,LeBitmap.Height-MargeB,LeBitmap.Width-MargeD,clBlack32);
  LeBitmap.VertLineS(LeBitmap.Width-MargeD,LeBitmap.Height-MargeB,MargeH,clBlack32);
  LeBitmap.HorzLineS(LeBitmap.Width-MargeD,MargeH,MargeG,clBlack32);
  // tracé des graduations
  LesGraduations := CalculeGraduations(LeBitmap,MaxiX,MaxiY,Param1,Param2,ValXEchLog,ValYEchLog);
  for i := 0 to Length(LesGraduations.PX)-1 do
    begin
      LeBitmap.VertLineS(Round(LesGraduations.PX[i]),LeBitmap.Height-MargeB,LeBitmap.Height-MargeB+5,clBlack32);
      lgt := LeBitmap.TextWidth(LesGraduations.VX[i])/2;
      XT := Round(LesGraduations.PX[i]-lgt);
      if (XT + 2*lgt) > (LeBitmap.Width - 2) then
        XT := LeBitmap.Width - 2 - Round(2*lgt);
      LeBitmap.RenderText(XT,LeBitmap.Height-MargeB+6,LesGraduations.VX[i],qltt,clBlack32);
    end;
  for i := 0 to Length(LesGraduations.PY)-1 do
    begin
      LeBitmap.HorzLineS(margeG,LesGraduations.PY[i],margeG-5,clBlack32);
      lgt := LeBitmap.TextWidth(LesGraduations.VY[i]);
      htt := LeBitmap.TextHeight(LesGraduations.VY[i])/2;
      LeBitmap.RenderText(Round(margeG-6-lgt),Round(LesGraduations.PY[i]-htt),LesGraduations.VY[i],qltt,clBlack32);
    end;
  LesGraduations.Free;
  // titres des axes
  lgt :=LeBitmap.TextWidth(TitreX);
  htt := LeBitmap.TextHeight(TitreX);
  LeBitmap.RenderText(Round(MargeG+(LeBitmap.Width - lgt)/2),Round(LeBitmap.Height-2-htt),TitreX,qltt,clBlack32);
  LeBitmap.RenderText(2,2,TitreY,qltt,clBlack32);
  if CoulDens then
    CalculeDensite(largeur, hauteur,NumParam1, NumParam2, MaXiY, dx, dy, Taillepoints,ValXEchLog,ValYEchLog);
  for i := 0 to NbEvenements - 1 do
    begin
      if Couleurs[i] <> clWhite32 then
        begin
          ValX := Donnees[i,NumParam1];
          ValY := Donnees[i,NumParam2];
          If (ValX >= 1) and (ValY >= 1) then
            begin
              if ValXEchLog then
                begin
                  if ValX <10 then
                    ValX := 10;
                  X := MargeG + Round(200*(log10(FracAleatoire(ValX))-1)*dx)
                end
              else
                X := MargeG+ Round(ValX*dX);
              XF := X+Round(dx*Taillepoints);
              if XF = X then
                XF := X+1;
              if ValYEchLog then
                begin
                  if ValY <10 then
                    ValY := 10;
                  Y := MargeH + Round((MaxiY - 200*(log10(ValY)-1))*dY)
                end
              else
                Y := MargeH + Round((MaxiY - ValY)*dY);
              YF := Y+Round(dy*Taillepoints);
              if YF = Y then
                YF := Y+1;
              LeBitmap.FillRectS(Rect(X,Y,XF,YF),Couleurs[i]);
            end;
        end;
    end;
  // Tracé des polygones  //** à ajuster suivant ValXEchLog,ValYEchLog
  for i := 0 to LesCategories.NbCategorieDeBase - 1 do
    begin
      MaCategorie := LesCategories.LesCategoriesDeBase[i];
      if (MaCategorie.Active or ((MaCategorie.CatFille<>nil) and MaCategorie.CatFille.Active) and (MaCategorie.NbParams = 2)) then
        begin
          if MaCategorie.Active then
            CouleurP := MaCategorie.Couleur
          else if (MaCategorie.CatFille<>nil) and MaCategorie.CatFille.Active then
            CouleurP := MaCategorie.CatFille.Couleur;
          if (MaCategorie.Params[0] = Param1) and (MaCategorie.Params[1] = Param2) then
            begin
              MonPolyg := TransposePolygoneInv(MaCategorie.PolygoneDef,LeBitmap.Width,LeBitmap.Height,MaxiX,MaxiY,False,ValXEchLog,ValYEchLog);
              try
                MonPolyg.DrawEdge(LeBitmap,CouleurP);
              finally
                MonPolyg.Free;
              end;
            end
          else if (MaCategorie.Params[1] = Param1) and (MaCategorie.Params[0] = Param2) then
            begin
              MonPolyg := TransposePolygoneInv(MaCategorie.PolygoneDef,LeBitmap.Width,LeBitmap.Height,MaxiX,MaxiY,True,ValXEchLog,ValYEchLog);
              try
                MonPolyg.DrawEdge(LeBitmap,CouleurP);
              finally
                MonPolyg.Free;
              end;
            end;
        end;
    end;
  LeBitmap.EndUpdate;
end;
procedure TDonneesCyto.RazNbParPixel;
begin
  NbParPixel.SetSize(0,0);
end;

procedure TDonneesCyto.SetLengthNbParPixel(Nblg,Nbht : integer);
begin
  NbParPixel.SetSize(Nblg,Nbht);
  NbParPixel.Clear(0);
end;

procedure TDonneesCyto.CalculeDensite(largeur, hauteur : integer; NumParam1, NumParam2 : integer; MaXiY, dx, dy : real; Taillepoints : integer;ValXEchLog,ValYEchLog : Boolean);
Var
  i, j : integer;
  ValX, ValY:real;
  X,Y,Xi,Yi : integer;
  Nblg : integer;
  Nbht : integer;
  H : single;
  ValD : integer;
begin
  Nblg := (largeur div Taillepoints) + 2;
  Nbht := (hauteur div Taillepoints) + 2;
  if NbParPixel = nil then
    NbParPixel := TBitmap32.Create;

  RazNbParPixel;
  SetLengthNbParPixel(Nblg,Nbht);
  NbParPixel.BeginUpdate;
  DensiteMax := 0;
  for i := 0 to NbEvenements - 1 do
    begin
      if Couleurs[i] <> clWhite32 then
        begin
          ValX := Donnees[i,NumParam1];
          ValY := Donnees[i,NumParam2];
          If (ValX >= 1) and (ValY >= 1) then
            begin
              if ValXEchLog then
                begin
                  if ValX <10 then
                    ValX := 10;
                  X := Round(200*(log10(FracAleatoire(ValX))-1)*dx)
                end
              else
                X := Round(ValX*dX);
              if ValYEchLog then
                begin
                  if ValY <10 then
                    ValY := 10;
                  Y := Round((MaxiY - 200*(log10(ValY)-1))*dY)
                end
              else
                Y := Round((MaxiY - ValY)*dY);
              Xi := X div Taillepoints;
              Yi := Y div TaillePoints;
              if (Xi < Nblg) and (Yi < Nbht) and (Xi > 0) and (Yi > 0) then
                begin
                  ValD :=NbParPixel[Xi,Yi];
                  inc(ValD);
                  NbParPixel[Xi,Yi]:=ValD;
                  if ValD > DensiteMax then
                    DensiteMax := ValD;
                end;
            end;
        end;
    end;
  NbParPixel.EndUpdate;
  // Dégradé de couleurs suivant la résolution
  for i := 0 to NbEvenements - 1 do
    begin
      if Couleurs[i] <> clWhite32 then
        begin
          ValX := Donnees[i,NumParam1];
          ValY := Donnees[i,NumParam2];
          If (ValX >= 1) and (ValY >= 1) then
            begin
              if ValXEchLog then
                begin
                  if ValX <10 then
                    ValX := 10;
                  X := Round(200*(log10(FracAleatoire(ValX))-1)*dx)
                end
              else
                X := Round(ValX*dX);
              if ValYEchLog then
                begin
                  if ValY <10 then
                    ValY := 10;
                  Y := Round((MaxiY - 200*(log10(ValY)-1))*dY)
                end
              else
                Y := Round((MaxiY - ValY)*dY);
              Xi := X div Taillepoints;
              Yi := Y div TaillePoints;
              if (Xi < NbParPixel.Width) and (Yi < NbParPixel.Height) and (Xi > 0) and (Yi > 0) and (NbParPixel[Xi,Yi] > 1) then
                begin
                  if DensiteMax >1 then
                    H := 0.7*(1-log10(NbParPixel[Xi,Yi])/log10(DensiteMax)) // échelle log
                  else
                    H := 0.7 *(1 - NbParPixel[Xi,Yi]/DensiteMax);
                  if H < 0 then H := 0;
                  Couleurs[i] := HSLtoRGB(H, 1, 0.5);
                end
              else
                Couleurs[i] := clBlack32;
            end;
        end;
    end;
end;

Constructor TDonneesCyto.Cree(Nm:string;Nombre,Nbp : integer);
Var
  i : integer;
begin
  Create;
  FacteurMult := 1;
  Nom := Nm;
  NbEvenements := Nombre;
  NbParam := Nbp;
  SetLength(Couleurs,NbEvenements);
  SetLength(Donnees,NbEvenements);
  for i := 0 to NbEvenements - 1 do
    begin
      SetLength(Donnees[i],NbParam);
      Couleurs[i] := $FF000000;
    end;
  NbParPixel := TBitmap32.Create;
end;
end.
