unit FMemo;
{$mode delphi}
interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ExtCtrls, Menus, StdCtrls;

type
  TFDonnees = class(TForm)
    Memo: TMemo;
    MainMenu1: TMainMenu;
    PChoix: TPanel;
    ComboFichier: TComboBox;
    Label1: TLabel;
    MFichier: TMenuItem;
    MEdition: TMenuItem;
    Menregistrer: TMenuItem;
    MFermer: TMenuItem;
    MToutCopier: TMenuItem;
    MCopierSelection: TMenuItem;
    SaveMemo: TSaveDialog;
    PPatienter: TPanel;
    procedure MFermerClick(Sender: TObject);
    procedure MFichierClick(Sender: TObject);
    procedure MCopierSelectionClick(Sender: TObject);
    procedure MToutCopierClick(Sender: TObject);
    procedure ComboFichierChange(Sender: TObject);
  private
    { Déclarations privées }
  public
    { Déclarations publiques }
    procedure RemplitMemo(ind : integer);
    procedure RemplitStringList(ind : integer);
  end;

var
  FDonnees: TFDonnees;

implementation

{$R *.lfm}
  uses Donnees;
procedure TFDonnees.ComboFichierChange(Sender: TObject);
Var
  Ind : integer;
  MesDonneesCyto : TDonneesCyto;
  MonMessage : string;
begin
  Ind := ComboFichier.ItemIndex;
  if Ind > -1 then
    begin
      MesDonneesCyto := LesDonneesCyto[ind];
      if MesDonneesCyto <> nil then
        begin
          if MesDonneesCyto.NbEvenements > 100000 then
            begin
              MonMessage := 'Il y a '+IntTostr(MesDonneesCyto.NbEvenements)+' données, l''affichage va prendre plusieurs minutes'+chr(10)+chr(13)+'Continuer ?';
              if MessageDlg(MonMessage,TMsgDlgType.mtConfirmation,[mbYes,mbNo],0) = mrNo then
                exit;
            end;
        end;
      PPatienter.Visible := True;
      Application.ProcessMessages;
      RemplitStringList(ind);
      PPatienter.Visible := False;
    end;
end;

procedure TFDonnees.MCopierSelectionClick(Sender: TObject);
Var
  MonMessage : string;
begin
  if Memo.SelLength > 0 then
    Memo.CopyToClipboard
  else
    begin
      MonMessage := 'Rien n''est sélectionné';
      MessageDlg(MonMessage,TMsgDlgType.mtError,[TMsgDlgBtn.mbOK],0);
    end;
end;

procedure TFDonnees.MFermerClick(Sender: TObject);
begin
 Close;
end;

procedure TFDonnees.MFichierClick(Sender: TObject);
begin
  if Memo.Lines.Count > 0 then
    begin
      if SaveMemo.Execute then
        Memo.Lines.SaveToFile(SaveMemo.FileName);
    end;
end;

procedure TFDonnees.MToutCopierClick(Sender: TObject);
begin
  Memo.SelectAll;
  Memo.CopyToClipboard;
  Memo.SelStart := 0;
  Memo.SelLength :=0;
end;
procedure TFDonnees.RemplitStringList(ind : integer);
Var
  i,j: integer;
  Chaine : String;
  MesDonneesCyto : TDonneesCyto;
  MaListe : TStringList;
begin
if (Length(LesDonneesCyto) > 0) and (ind > -1) and (ind < Length(LesDonneesCyto)) then
  begin
    MesDonneesCyto := LesDonneesCyto[ind];
    MaListe := TStringList.Create;
    try
      MaListe.Add(MesDonneesCyto.LaDate);
      MaListe.Add('Cytomètre : '+ MesDonneesCyto.Cytometre);
      MaListe.Add(MesDonneesCyto.FormatFichier);
      MaListe.Add('Nombre de paramètres : '+IntToStr(MesDonneesCyto.NbParam));
      For i :=0 to MesDonneesCyto.NbParam -1 do
        begin
          Chaine := 'Paramètre '+IntToStr(i+1)+' : '+MesDonneesCyto.Parametres[i].Nom
              +' Alias : '+ MesDonneesCyto.Parametres[i].Alias;
          MaListe.Add(chaine);
          Chaine := '  Résolution : '+IntToStr(MesDonneesCyto.Parametres[i].Resolution)
              +' Nombre de bits : '+ IntToStr(MesDonneesCyto.Parametres[i].NombreBits)+'   ';
          if MesDonneesCyto.Parametres[i].TypeEchelle = teLin then
            Chaine := Chaine + 'Lin   '
          else
            Chaine := Chaine + 'Log   ';
          Chaine := Chaine + FloatToStr(MesDonneesCyto.Parametres[i].Puis10Log);
          MaListe.Add(Chaine);
        end;
      MaListe.Add('Nombre d''événements : '+IntToStr(MesDonneesCyto.NbEvenements));
      for i :=0 to MesDonneesCyto.NbEvenements-1 do
        begin
          Chaine :='';
          for j :=0  to MesDonneesCyto.NbParam -1 do
            chaine := Chaine + '   '+FloatToStr(MesDonneesCyto.Donnees[i,j]);
          MaListe.Add(Chaine);
        end;
      Memo.Lines.AddStrings(MaListe);
    finally
      MaListe.Free;
    end;
  end;
end;

procedure TFDonnees.RemplitMemo(ind : integer);
Var
  i,j: integer;
  Chaine : String;
  MesDonneesCyto : TDonneesCyto;
begin
if (Length(LesDonneesCyto) > 0) and (ind > -1) and (ind < Length(LesDonneesCyto)) then
  begin
    MesDonneesCyto := LesDonneesCyto[ind];
    Memo.Lines.Add(MesDonneesCyto.LaDate);
    Memo.Lines.Add(MesDonneesCyto.FormatFichier);
    if MesDonneesCyto.Projet <> '' then
      Memo.Lines.Add('Nom du projet : '+ MesDonneesCyto.Projet);
    if MesDonneesCyto.Cellules <> '' then
      Memo.Lines.Add('Cellules étudiées : '+ MesDonneesCyto.Cellules);
    Memo.Lines.Add('Nombre de paramètres : '+IntToStr(MesDonneesCyto.NbParam));
    For i :=0 to MesDonneesCyto.NbParam -1 do
      begin
        Chaine := 'Paramètre '+IntToStr(i+1)+' : '+MesDonneesCyto.Parametres[i].Nom
            +' Alias : '+ MesDonneesCyto.Parametres[i].Alias;
        Memo.Lines.Add(chaine);
        Chaine := '  Résolution : '+IntToStr(MesDonneesCyto.Parametres[i].Resolution)
            +' Nombre de bits : '+ IntToStr(MesDonneesCyto.Parametres[i].NombreBits)+'   ';
        if MesDonneesCyto.Parametres[i].TypeEchelle = teLin then
          Chaine := Chaine + 'Lin   '
        else
          Chaine := Chaine + 'Log   ';
        Chaine := Chaine + FloatToStr(MesDonneesCyto.Parametres[i].Puis10Log);
        Memo.Lines.Add(Chaine);
      end;
    Memo.Lines.Add('Nombre d''événements : '+IntToStr(MesDonneesCyto.NbEvenements));
    for i :=0 to MesDonneesCyto.NbEvenements-1 do
      begin
        Chaine :='';
        for j :=0  to MesDonneesCyto.NbParam -1 do
          chaine := Chaine + '   '+FloatToStr(MesDonneesCyto.Donnees[i,j]);
        Memo.Lines.Add(Chaine);
      end;
  end;
end;
end.
