unit formtableau;

{$mode delphi}

interface

uses
  Classes, SysUtils, FileUtil, TAGraph, TASeries, TASources, TALegendPanel,
  Forms, Controls, Graphics, Dialogs, ExtCtrls, Grids, StdCtrls, Menus, ExtDlgs;

type

  { TFTableau }

  TFTableau = class(TForm)
    BOKTitre: TButton;
    Chart1: TChart;
    EdiTitre: TEdit;
    MainMenu1: TMainMenu;
    MCategories: TMenuItem;
    MEdition: TMenuItem;
    MCopierTableau: TMenuItem;
    MCopierGraphe: TMenuItem;
    MEnregistrerIm: TMenuItem;
    MTitre: TMenuItem;
    MenuItem2: TMenuItem;
    MImprim: TMenuItem;
    MFermer: TMenuItem;
    MFichier: TMenuItem;
    PanelBas: TPanel;
    PanelTab: TPanel;
    PanelGraphe: TPanel;
    PChoixTitre: TPanel;
    SaveDialogIm: TSaveDialog;
    Splitter1: TSplitter;
    Tableau: TStringGrid;
    procedure BOKTitreClick(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure MCopierGrapheClick(Sender: TObject);
    procedure MCopierTableauClick(Sender: TObject);
    procedure MEnregistrerImClick(Sender: TObject);
    procedure MFermerClick(Sender: TObject);
    procedure MImprimClick(Sender: TObject);
    procedure MTitreClick(Sender: TObject);
  private
    { private declarations }
  public
    { public declarations }
    numCatRestr : integer;
    procedure Afficher;
    procedure MCategorieClick(Sender: TObject);
  end;

var
  FTableau: TFTableau;

implementation

{$R *.lfm}
uses donnees,Clipbrd,GR32,Gestimpr;

procedure TFTableau.MFermerClick(Sender: TObject);
begin
  Close;
end;
procedure TFTableau.MCategorieClick(Sender: TObject);
Var
  UnMenuItem : TMenuItem;
begin
  if Sender is TMenuItem then
    begin
      UnMenuItem := (Sender as TMenuItem);
      if UnMenuItem <> nil then
        begin
          UnMenuItem.Checked:=Not UnMenuItem.Checked;
          Afficher;
        end;
    end;
end;

procedure TFTableau.MImprimClick(Sender: TObject);
Var
  BitmapTemp : TBitmap32;
begin
  BitmapTemp := TBitmap32.Create;
  try
    BitmapTemp.Width:=Chart1.Width;
    BitmapTemp.Height:=Chart1.Height;
    Chart1.DrawOnCanvas(Rect(0,0,Chart1.Width,Chart1.Height),BitmapTemp.Canvas);
    FGestImp.Imprime(BitmapTemp);
  finally
    Bitmaptemp.Free;
  end;
end;

procedure TFTableau.MTitreClick(Sender: TObject);
begin
  PChoixTitre.Visible:=True;
end;


procedure TFTableau.MCopierGrapheClick(Sender: TObject);
begin
  Chart1.CopyToClipboardBitmap;
end;

procedure TFTableau.MCopierTableauClick(Sender: TObject);
begin
  Tableau.CopyToClipboard(False);
end;

procedure TFTableau.MEnregistrerImClick(Sender: TObject);
begin
  if SaveDialogIm.Execute then
    Chart1.SaveToBitmapFile(SaveDialogIm.FileName);
end;

procedure TFTableau.FormCreate(Sender: TObject);
begin
  numCatRestr := -1;
end;

procedure TFTableau.BOKTitreClick(Sender: TObject);
begin
  if Trim(EdiTitre.Text) <> '' then
    begin
      Chart1.Title.Text.Text :=EdiTitre.Text;
      Chart1.Title.Visible:=True;
    end;
  PChoixTitre.Visible:=False;
end;

procedure TFTableau.Afficher;
Var
  i,j,col,numc, NbRef :integer;
  ValY : real;
  UneSerie : TLineSeries;
  chaine, pourcent, titreS : string;
begin
  MCategories.Visible:=(MCategories.Count > 1);
  Chart1.ClearSeries; // supprimer toutes les séries
  Tableau.Cells[0,0] := LesDonneesCyto[0].ParamSerie+' ('+LesDonneesCyto[0].UniteParamSerie+')';
  Tableau.ColWidths[0]:=Tableau.Canvas.TextWidth(LesDonneesCyto[0].ParamSerie+' ('+LesDonneesCyto[0].UniteParamSerie+')')+8;
  Tableau.RowCount := Length(LesDonneesCyto) + 1;
  Chart1.BottomAxis.Title.Caption:=Tableau.Cells[0,0];
  for i := 0 to Length(LesDonneesCyto) - 1 do
    begin
      Tableau.Cells[0,i+1] := IntToStr(LesDonneesCyto[i].ValeurParamSerie);
    end;
  col := 0;
  for j := 0 to MCategories.Count -1 do
    begin
      if MCategories.Items[j].Checked then
        begin
          inc(col);
          if Tableau.ColCount<=col then
            Tableau.ColCount:=col+1;
          Tableau.Cells[col,0]:=MCategories.Items[j].Caption+' (%)';
          Tableau.ColWidths[col]:=Tableau.Canvas.TextWidth(MCategories.Items[j].Caption+' (%)')+8;
          //créer une série pour chaque courbe à tracer
          UneSerie := TLineSeries.Create(Chart1);
          titreS := MCategories.Items[j].Caption;
          UneSerie.Title := titreS;
          Chart1.AddSeries(UneSerie);
          if MCategories.Items[j].Tag <0 then
            begin
              numc := -2-MCategories.Items[j].Tag;
              UneSerie.LinePen.Color:=Wincolor(LesCategories.LesCategoriesComp[numc].Couleur);
              if numCatRestr = -1 then
                begin
                  PanelBas.Caption:='';
                  Pourcent := '(%)';
                  for i := 0 to Length(LesDonneesCyto) - 1 do
                    begin
                      ValY := 100*(LesDonneesCyto[i].NbEventCC[numc]/LesDonneesCyto[i].NbEvenements);
                      Tableau.Cells[col,i+1]:=FloatToStrF(ValY,ffFixed,8,2);
                    end;
                end
              else
                begin
                  for i := 0 to Length(LesDonneesCyto) - 1 do
                    begin
                      if numCatRestr > -1 then
                        begin
                          NbRef := LesDonneesCyto[i].NbEventCS[numCatRestr];
                          PanelBas.Caption:='% calculés par rapport à la catégorie '
                            +LesCategories.LesCategoriesDeBase[numCatRestr].Nom;
                          Pourcent := '(% des '+lowercase(LesCategories.LesCategoriesDeBase[numCatRestr].Nom)+')';
                        end
                      else
                        begin
                          NbRef := LesDonneesCyto[i].NbEventCC[-2-numCatRestr];
                          PanelBas.Caption:='% calculés par rapport à la catégorie '
                            +LesCategories.LesCategoriesComp[-2-numCatRestr].Nom;
                          Pourcent := '(% des '+lowercase(LesCategories.LesCategoriesComp[-2-numCatRestr].Nom)+')';
                        end;

                      if NbRef > 0 then
                        ValY := 100*(LesDonneesCyto[i].NbEventCC[numc]/NbRef)
                      else
                        ValY := 0;
                      Tableau.Cells[col,i+1]:=FloatToStrF(ValY,ffFixed,8,2);
                      UneSerie.AddXY(LesDonneesCyto[i].ValeurParamSerie,ValY);
                    end;
                end;
            end
          else
            begin
              numc := MCategories.Items[j].Tag;
              for i := 0 to Length(LesDonneesCyto) - 1 do
                begin
                  UneSerie.LinePen.Color:=Wincolor(LesCategories.LesCategoriesDeBase[numc].Couleur);
                  ValY := 100*(LesDonneesCyto[i].NbEventCS[numc]/LesDonneesCyto[i].NbEvenements);
                  Tableau.Cells[col,i+1]:=FloatToStrF(ValY,ffFixed,8,2);
                  UneSerie.AddXY(LesDonneesCyto[i].ValeurParamSerie,ValY);
                end;
            end;
        end;
      if col = 1 then
        Chart1.LeftAxis.Title.Caption:= titreS+' '+ Pourcent //Tableau.Cells[1,0]
      else
        begin
          Chart1.Legend.Visible:=True;
          Chart1.LeftAxis.Title.Caption:= Pourcent;
        end;
    end;
end;
end.

